# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
The cadcutils.util package contains utilities commonly use in CADC
applications.

The following functions are meant to be imported directly from cadcutils.util:

    - date2ivoa, str2ivoa: conversion of date to/from string
    - get_logger: returns a logger that logs in CADC format
    - get_log_level: returns the logger level
    - get_base_parser: creates a basic parser for CADC web app applications

"""
from .utils import *  # noqa
from .config import *  # noqa



# -*- coding: utf-8 -*-
# ***********************************************************************
# ******************  CANADIAN ASTRONOMY DATA CENTRE  *******************
# *************  CENTRE CANADIEN DE DONNÉES ASTRONOMIQUES  **************
#
#  (c) 2023.                            (c) 2023.
#  Government of Canada                 Gouvernement du Canada
#  National Research Council            Conseil national de recherches
#  Ottawa, Canada, K1A 0R6              Ottawa, Canada, K1A 0R6
#  All rights reserved                  Tous droits réservés
#
#  NRC disclaims any warranties,        Le CNRC dénie toute garantie
#  expressed, implied, or               énoncée, implicite ou légale,
#  statutory, of any kind with          de quelque nature que ce
#  respect to the software,             soit, concernant le logiciel,
#  including without limitation         y compris sans restriction
#  any warranty of merchantability      toute garantie de valeur
#  or fitness for a particular          marchande ou de pertinence
#  purpose. NRC shall not be            pour un usage particulier.
#  liable in any event for any          Le CNRC ne pourra en aucun cas
#  damages, whether direct or           être tenu responsable de tout
#  indirect, special or general,        dommage, direct ou indirect,
#  consequential or incidental,         particulier ou général,
#  arising from the use of the          accessoire ou fortuit, résultant
#  software.  Neither the name          de l'utilisation du logiciel. Ni
#  of the National Research             le nom du Conseil National de
#  Council of Canada nor the            Recherches du Canada ni les noms
#  names of its contributors may        de ses  participants ne peuvent
#  be used to endorse or promote        être utilisés pour approuver ou
#  products derived from this           promouvoir les produits dérivés
#  software without specific prior      de ce logiciel sans autorisation
#  written permission.                  préalable et particulière
#                                       par écrit.
#
#  This file is part of the             Ce fichier fait partie du projet
#  OpenCADC project.                    OpenCADC.
#
#  OpenCADC is free software:           OpenCADC est un logiciel libre ;
#  you can redistribute it and/or       vous pouvez le redistribuer ou le
#  modify it under the terms of         modifier suivant les termes de
#  the GNU Affero General Public        la “GNU Affero General Public
#  License as published by the          License” telle que publiée
#  Free Software Foundation,            par la Free Software Foundation
#  either version 3 of the              : soit la version 3 de cette
#  License, or (at your option)         licence, soit (à votre gré)
#  any later version.                   toute version ultérieure.
#
#  OpenCADC is distributed in the       OpenCADC est distribué
#  hope that it will be useful,         dans l’espoir qu’il vous
#  but WITHOUT ANY WARRANTY;            sera utile, mais SANS AUCUNE
#  without even the implied             GARANTIE : sans même la garantie
#  warranty of MERCHANTABILITY          implicite de COMMERCIALISABILITÉ
#  or FITNESS FOR A PARTICULAR          ni d’ADÉQUATION À UN OBJECTIF
#  PURPOSE.  See the GNU Affero         PARTICULIER. Consultez la Licence
#  General Public License for           Générale Publique GNU Affero
#  more details.                        pour plus de détails.
#
#  You should have received             Vous devriez avoir reçu une
#  a copy of the GNU Affero             copie de la Licence Générale
#  General Public License along         Publique GNU Affero avec
#  with OpenCADC.  If not, see          OpenCADC ; si ce n’est
#  <http://www.gnu.org/licenses/>.      pas le cas, consultez :
#                                       <http://www.gnu.org/licenses/>.
#
#
# ***********************************************************************

import os
import sys
import re
import unittest

from unittest.mock import Mock, patch, mock_open
from io import StringIO

from cadcutils.net import auth
from cadcutils.util import get_base_parser

THIS_DIR = os.path.dirname(os.path.realpath(__file__))
TESTDATA_DIR = os.path.join(THIS_DIR, 'data')


class MyExitError(Exception):
    pass


class TestAuth(unittest.TestCase):
    """ Class for testing networking authorization functionality """

    @patch('cadcutils.net.auth.get_cert', Mock(return_value='CERTVALUE'))
    @patch('cadcutils.net.auth.util.check_version', Mock())  # mock from get_cert function
    @patch('cadcutils.util.utils.check_version', Mock())  # mock from parse args
    @patch('sys.exit', Mock(side_effect=[MyExitError, MyExitError]))
    def test_get_cert_main(self):
        """ Test the cert_main function """

        value = "CERTVALUE"

        # get certificate with no arguments -> authentication args required
        with self.assertRaises(MyExitError):
            sys.argv = ["cadc-get-cert"]
            auth.get_cert_main()

        # get certificate default location
        m = mock_open()
        with patch('builtins.open', m, create=True):
            sys.argv = ["cadc-get-cert", "-u", "bob"]
            auth.get_cert_main()
        m.assert_called_with(
            os.path.join(os.getenv('HOME', '/tmp'), '.ssl/cadcproxy.pem'), 'w')
        handle = m()
        handle.write.assert_called_once_with(value)

        # save certificate in a file
        certfile = '/tmp/testcertfile'
        try:
            os.remove(certfile)
        except OSError:
            pass
        sys.argv = ["cadc-get-cert", "-u", "bob", "--cert-filename", certfile]
        self.assertEqual(None, auth.get_cert_main())
        with open(certfile, 'r') as f:
            self.assertEqual(value, f.read())

        # test error when the directory of the cert file is in fact an
        # existing file..
        errmsg = """[Errno 17] File exists: '{}' : {}
Expected /tmp/testcertfile to be a directory.
""".format(certfile, certfile)
        sys.argv = ["cadc-get-cert", "-u", "bob", "--cert-filename",
                    certfile + "/cert"]
        with self.assertRaises(MyExitError):
            with patch('sys.stderr', new_callable=StringIO) as stderr_mock:
                auth.get_cert_main()
        self.assertEqual(errmsg, stderr_mock.getvalue())
        os.remove(certfile)

    @patch('sys.exit', Mock(side_effect=[MyExitError]))
    def test_get_cert_main_help(self):
        """ Test the help option of the cadc-get-cert app """
        with open(os.path.join(TESTDATA_DIR, 'help_cadc-get-cert.txt'),
                  'r') as f:
            usage = f.read()
        # update the default cert location line
        usage = re.sub(
            r'default: .*cadcproxy.pem',
            'default: {}/.ssl/cadcproxy.pem'.format(os.getenv("HOME")),
            usage)

        # --help
        self.maxDiff = None  # Display the entire difference
        with patch('sys.stdout', new_callable=StringIO) as stdout_mock:
            sys.argv = ["cadc-get-cert", "--help"]
            with self.assertRaises(MyExitError):
                auth.get_cert_main()
            # new title in 3.10
            actual = stdout_mock.getvalue().replace('options:',
                                                    'optional arguments:')
            self.assertEqual(usage.strip('\n'), actual.strip('\n'))

    @patch('cadcutils.net.auth.os')
    def testSubject(self, os_mock):
        # anon subject
        subject = auth.Subject()
        self.assertTrue(subject.anon)
        self.assertEqual(None, subject.certificate)
        self.assertEqual({}, subject._hosts_auth)
        self.assertEqual(None, subject.get_auth('realm1'))

        # cert subject
        cert = 'somecert'
        subject = auth.Subject(certificate=cert)
        self.assertFalse(subject.anon)
        self.assertEqual(cert, subject.certificate)
        self.assertEqual({}, subject._hosts_auth)
        self.assertEqual(None, subject.get_auth('realm1'))

        # empty netrc subject
        m = mock_open()
        with patch('builtins.open', m, create=True):
            subject = auth.Subject(netrc='somefile')
        self.assertFalse(subject.anon)
        self.assertEqual(None, subject.certificate)
        self.assertEqual({}, subject._hosts_auth)
        self.assertEqual(None, subject.get_auth('realm1'))

        # netrc with content
        netrc_content = {'realm1': ('user1', None, 'pass1'),
                         'realm2': ('user1', None, 'pass2')}
        expected_host_auth = {'realm1': ('user1', 'pass1'),
                              'realm2': ('user1', 'pass2')}
        os_mock.path.join.return_value = '/home/myhome/.netrc'
        with patch('cadcutils.net.auth.netrclib') as netrclib_mock:
            netrclib_mock.netrc.return_value.hosts = netrc_content
            subject = auth.Subject(netrc=True)
        self.assertFalse(subject.anon)
        self.assertEqual(None, subject.certificate)
        self.assertEqual('/home/myhome/.netrc', subject.netrc)
        self.assertEqual(expected_host_auth, subject._hosts_auth)
        self.assertEqual(('user1', 'pass1'), subject.get_auth('realm1'))
        self.assertEqual(('user1', 'pass2'), subject.get_auth('realm2'))
        self.assertEqual(None, subject.get_auth('realm3'))

        # subject with username
        username = 'user1'
        passwd = 'passwd1'
        subject = auth.Subject(username=username)
        self.assertFalse(subject.anon)
        self.assertEqual(None, subject.certificate)
        self.assertEqual({}, subject._hosts_auth)
        with patch('cadcutils.net.auth.getpass') as getpass_mock:
            getpass_mock.getpass.return_value = passwd
            self.assertEqual((username, passwd), subject.get_auth('realm1'))

        parser = get_base_parser(subparsers=False)
        args = parser.parse_args(['--resource-id', 'blah'])
        subject = auth.Subject.from_cmd_line_args(args)
        self.assertTrue(subject.anon)

        sys.argv = ['cadc-client', '--resource-id', 'blah', '--cert',
                    'mycert.pem']
        args = parser.parse_args()
        subject = auth.Subject.from_cmd_line_args(args)
        self.assertFalse(subject.anon)
        self.assertEqual('mycert.pem', subject.certificate)

        # subject with cookies
        cookie = auth.CookieInfo('some.domain', 'MyCookie', 'somevalue')
        subject = auth.Subject()
        subject.cookies.append(cookie)
        assert auth.SECURITY_METHODS_IDS['cookie'] in\
            subject.get_security_methods()



# -*- coding: utf-8 -*-
# ***********************************************************************
# ******************  CANADIAN ASTRONOMY DATA CENTRE  *******************
# *************  CENTRE CANADIEN DE DONNÉES ASTRONOMIQUES  **************
#
# (c) 2021.                            (c) 2021.
#  Government of Canada                 Gouvernement du Canada
#  National Research Council            Conseil national de recherches
#  Ottawa, Canada, K1A 0R6              Ottawa, Canada, K1A 0R6
#  All rights reserved                  Tous droits réservés
#
#  NRC disclaims any warranties,        Le CNRC dénie toute garantie
#  expressed, implied, or               énoncée, implicite ou légale,
#  statutory, of any kind with          de quelque nature que ce
#  respect to the software,             soit, concernant le logiciel,
#  including without limitation         y compris sans restriction
#  any warranty of merchantability      toute garantie de valeur
#  or fitness for a particular          marchande ou de pertinence
#  purpose. NRC shall not be            pour un usage particulier.
#  liable in any event for any          Le CNRC ne pourra en aucun cas
#  damages, whether direct or           être tenu responsable de tout
#  indirect, special or general,        dommage, direct ou indirect,
#  consequential or incidental,         particulier ou général,
#  arising from the use of the          accessoire ou fortuit, résultant
#  software.  Neither the name          de l'utilisation du logiciel. Ni
#  of the National Research             le nom du Conseil National de
#  Council of Canada nor the            Recherches du Canada ni les noms
#  names of its contributors may        de ses  participants ne peuvent
#  be used to endorse or promote        être utilisés pour approuver ou
#  products derived from this           promouvoir les produits dérivés
#  software without specific prior      de ce logiciel sans autorisation
#  written permission.                  préalable et particulière
#                                       par écrit.
#
#  This file is part of the             Ce fichier fait partie du projet
#  OpenCADC project.                    OpenCADC.
#
#  OpenCADC is free software:           OpenCADC est un logiciel libre ;
#  you can redistribute it and/or       vous pouvez le redistribuer ou le
#  modify it under the terms of         modifier suivant les termes de
#  the GNU Affero General Public        la “GNU Affero General Public
#  License as published by the          License” telle que publiée
#  Free Software Foundation,            par la Free Software Foundation
#  either version 3 of the              : soit la version 3 de cette
#  License, or (at your option)         licence, soit (à votre gré)
#  any later version.                   toute version ultérieure.
#
#  OpenCADC is distributed in the       OpenCADC est distribué
#  hope that it will be useful,         dans l’espoir qu’il vous
#  but WITHOUT ANY WARRANTY;            sera utile, mais SANS AUCUNE
#  without even the implied             GARANTIE : sans même la garantie
#  warranty of MERCHANTABILITY          implicite de COMMERCIALISABILITÉ
#  or FITNESS FOR A PARTICULAR          ni d’ADÉQUATION À UN OBJECTIF
#  PURPOSE.  See the GNU Affero         PARTICULIER. Consultez la Licence
#  General Public License for           Générale Publique GNU Affero
#  more details.                        pour plus de détails.
#
#  You should have received             Vous devriez avoir reçu une
#  a copy of the GNU Affero             copie de la Licence Générale
#  General Public License along         Publique GNU Affero avec
#  with OpenCADC.  If not, see          OpenCADC ; si ce n’est
#  <http://www.gnu.org/licenses/>.      pas le cas, consultez :
#                                       <http://www.gnu.org/licenses/>.
#
# ***********************************************************************

from lxml import etree
from urllib.parse import urlparse

from cadcutils.util import str2ivoa
from .. import Group
from .group_property_reader import GroupPropertyReader
from .user_reader import UserReader

GROUP_URI = 'ivo://cadc.nrc.ca/gms?'


class GroupReader(object):
    """GroupReader """

    def read(self, xml_string, deep_copy=True):
        """Build an Group object from an XML document string.

        Arguments:
        xml_string -- string of XML containing the Group element
        return     -- a Group object
        """
        root = etree.fromstring(xml_string, etree.XMLParser())
        return self.get_group(root, deep_copy)

    def get_group(self, group_element, deep_copy):
        uri = group_element.get("uri")
        if uri is None:
            raise GroupParsingException("group missing required uri attribute")
        if GROUP_URI not in uri:
            raise GroupParsingException(
                "group uri attribute malformed: {0}".format(uri))

        url_parts = urlparse(uri)
        authority = '{}://{}{}'.format(
            url_parts.scheme, url_parts.netloc, url_parts.path)
        group_id = url_parts.query
        group = Group(group_id=group_id, authority=authority)
        group.owner = self._get_owner(group_element)

        if deep_copy:
            group.description = self._get_child_text(group_element,
                                                     'description')
            last_modified = self._get_child_text(group_element, 'lastModified')
            if last_modified is not None:
                group.last_modified = str2ivoa(last_modified)
            group.properties = self._get_group_properties(group_element)
            group.group_members = self._get_groups(group_element,
                                                   'groupMembers')
            group.user_members = self._get_users(group_element, 'userMembers')
            group.group_admins = self._get_groups(group_element, 'groupAdmins')
            group.user_admins = self._get_users(group_element, 'userAdmins')

        return group

    def _get_child_text(self, parent, tag):
        child_element = self._get_child_element(parent, tag)
        if child_element is None:
            return None
        else:
            return child_element.text

    def _get_child_element(self, parent, tag):
        for element in list(parent):
            if element.tag == tag:
                if element.keys() or element.text:
                    # element has attributes or content, return it
                    return element
                break
        return None

    def _get_owner(self, group_element):
        user_element = group_element.find('./owner/user')
        if user_element is None:
            return None

        user_reader = UserReader()
        user = user_reader.get_user(user_element)
        return user

    def _get_group_properties(self, group_element):
        properties = set()
        properties_element = self._get_child_element(group_element,
                                                     'properties')
        if properties_element is not None:
            property_reader = GroupPropertyReader()
            for property_element in properties_element.findall('property'):
                properties.add(
                    property_reader.get_group_property(property_element))
        return properties

    def _get_groups(self, parent, tag):
        groups = set()
        group_elements = self._get_child_element(parent, tag)
        if group_elements is not None:
            for group in group_elements.findall('group'):
                groups.add(self.get_group(group, False))
        return groups

    def _get_users(self, parent, tag):
        users = set()
        user_elements = self._get_child_element(parent, tag)
        if user_elements is not None:
            for user in user_elements.findall('user'):
                user_reader = UserReader()
                user = user_reader.get_user(user)
                users.add(user)
        return users


class GroupParsingException(Exception):
    """A group exception class for catching XML parsing exception"""

    def __init__(self, value):
        self.value = value

    def __str__(self):
        return repr(self.value)



# -*- coding: utf-8 -*-

# ***********************************************************************
# ******************  CANADIAN ASTRONOMY DATA CENTRE  *******************
# *************  CENTRE CANADIEN DE DONNÉES ASTRONOMIQUES  **************
#
#  (c) 2022.                            (c) 2022.
#  Government of Canada                 Gouvernement du Canada
#  National Research Council            Conseil national de recherches
#  Ottawa, Canada, K1A 0R6              Ottawa, Canada, K1A 0R6
#  All rights reserved                  Tous droits réservés
#
#  NRC disclaims any warranties,        Le CNRC dénie toute garantie
#  expressed, implied, or               énoncée, implicite ou légale,
#  statutory, of any kind with          de quelque nature que ce
#  respect to the software,             soit, concernant le logiciel,
#  including without limitation         y compris sans restriction
#  any warranty of merchantability      toute garantie de valeur
#  or fitness for a particular          marchande ou de pertinence
#  purpose. NRC shall not be            pour un usage particulier.
#  liable in any event for any          Le CNRC ne pourra en aucun cas
#  damages, whether direct or           être tenu responsable de tout
#  indirect, special or general,        dommage, direct ou indirect,
#  consequential or incidental,         particulier ou général,
#  arising from the use of the          accessoire ou fortuit, résultant
#  software.  Neither the name          de l'utilisation du logiciel. Ni
#  of the National Research             le nom du Conseil National de
#  Council of Canada nor the            Recherches du Canada ni les noms
#  names of its contributors may        de ses  participants ne peuvent
#  be used to endorse or promote        être utilisés pour approuver ou
#  products derived from this           promouvoir les produits dérivés
#  software without specific prior      de ce logiciel sans autorisation
#  written permission.                  préalable et particulière
#                                       par écrit.
#
#  This file is part of the             Ce fichier fait partie du projet
#  OpenCADC project.                    OpenCADC.
#
#  OpenCADC is free software:           OpenCADC est un logiciel libre ;
#  you can redistribute it and/or       vous pouvez le redistribuer ou le
#  modify it under the terms of         modifier suivant les termes de
#  the GNU Affero General Public        la “GNU Affero General Public
#  License as published by the          License” telle que publiée
#  Free Software Foundation,            par la Free Software Foundation
#  either version 3 of the              : soit la version 3 de cette
#  License, or (at your option)         licence, soit (à votre gré)
#  any later version.                   toute version ultérieure.
#
#  OpenCADC is distributed in the       OpenCADC est distribué
#  hope that it will be useful,         dans l’espoir qu’il vous
#  but WITHOUT ANY WARRANTY;            sera utile, mais SANS AUCUNE
#  without even the implied             GARANTIE : sans même la garantie
#  warranty of MERCHANTABILITY          implicite de COMMERCIALISABILITÉ
#  or FITNESS FOR A PARTICULAR          ni d’ADÉQUATION À UN OBJECTIF
#  PURPOSE.  See the GNU Affero         PARTICULIER. Consultez la Licence
#  General Public License for           Générale Publique GNU Affero
#  more details.                        pour plus de détails.
#
#  You should have received             Vous devriez avoir reçu une
#  a copy of the GNU Affero             copie de la Licence Générale
#  General Public License along         Publique GNU Affero avec
#  with OpenCADC.  If not, see          OpenCADC ; si ce n’est
#  <http://www.gnu.org/licenses/>.      pas le cas, consultez :
#                                       <http://www.gnu.org/licenses/>.
#
#
# ***********************************************************************

"""
Contains functionality related to interacting with Web Services. Users
of this class can instantiate the BaseWsClient in order to access
the webservices via one of the get, put, post, delete and head
functions that the service supports.

"""

import logging
import sys
import time
import platform
import os
import hashlib

import requests
from requests import Session
from urllib.parse import urlparse
import distro

from cadcutils import exceptions, util, net
from cadcutils import version as cadctools_version
from . import wscapabilities

__all__ = ['BaseWsClient', 'BaseDataClient', 'get_resources', 'list_resources',
           'DEFAULT_REGISTRY']

BUFSIZE = 8388608  # Size of read/write buffer
MAX_RETRY_DELAY = 128  # maximum delay between retries
# start delay between retries when Try_After not sent by server.
DEFAULT_RETRY_DELAY = 30
MAX_NUM_RETRIES = 6

SERVICE_RETRY = 'Retry-After'
SERVICE_AVAILABILITY_ID = 'ivo://ivoa.net/std/VOSI#availability'

# Files up to this size might have their md5 checksum pre-computed before
# transferring. For larger files, the added overhead does not justify it.
# Can be overriden
MAX_MD5_COMPUTE_SIZE = 5 * 1024 * 1024
# can be overriden by environment
if os.getenv('CADC_MAX_MD5_COMPUTE_SIZE', None):
    MAX_MD5_COMPUTE_SIZE = int(os.getenv('CADC_MAX_MD5_COMPUTE_SIZE'))

# Files smaller that this size can be send with one request. Larger ones
# required to be split into segments that are sent and acknowledged
# individually
GIB = 1024 * 1024 * 1024
FILE_SEGMENT_THRESHOLD = 5 * GIB  # large files require segments
PREFERRED_SEGMENT_SIZE = 2 * GIB

MD5_MISMATCH_RETRY = 3  # number of times to retry on md5 mismatch errors

# HTTP attribute names
HTTP_LENGTH = 'Content-Length'

# PUT transactions headers/values
PUT_TXN_OP = 'x-put-txn-op'
PUT_TXN_ID = 'x-put-txn-id'
PUT_TXN_TOTAL_LENGTH = 'x-total-length'
PUT_TXN_MIN_SEGMENT = 'x-put-segment-minbytes'
PUT_TXN_MAX_SEGMENT = 'x-put-segment-maxbytes'
# transaction operations
PUT_TXN_START = 'start'
PUT_TXN_COMMIT = 'commit'
PUT_TXN_ABORT = 'abort'
PUT_TXN_REVERT = 'revert'

HEADERS = 'headers'  # name of the kwargs headers arg

# size of the read blocks in data transfers
READ_BLOCK_SIZE = 8 * 1024

# try to disable the unverified HTTPS call warnings
try:
    requests.packages.urllib3.disable_warnings()
except Exception:
    pass


def get_resources(with_caps=True):
    """
    Fetches the registry information regarding the available resources
    and their capabilities
    :param with_caps: True for including capabilities information,
    false otherwise
    :return: List of the form (resource_id, capability url,
    wscapabilities.Capabilities)}
    """
    resources = []
    cr = wscapabilities.CapabilitiesReader()
    response = requests.get(DEFAULT_REGISTRY)
    response.raise_for_status()
    for line in response.text.split('\n'):
        caps = None
        if (not line.startswith('#')) and (len(line.strip()) > 0):
            resource_id, url = line.split('=')
            if with_caps:
                caps_resp = requests.get(url.strip())
                caps_resp.raise_for_status()
                caps = cr.parsexml(caps_resp.text)
            resources.append((resource_id.strip(), url.strip(), caps))
    return resources


def list_resources():
    """
    Displays information about the services and their capabilities.
    """
    resources = get_resources()
    for r in resources:
        if r[2] is None:
            caps_str = 'NA'
        else:
            caps_str = ', '.join(sorted(r[2]._caps.keys()))
        print('{} ({}) - Capabilities: {}\n'.format(r[0], r[1], caps_str))


class BaseWsClient(object):
    """
    Web Service client primarily for CADC services. It is a wrapper class
       around the requests module to facilitate the interaction with the
       server via the get, put, post, delete and head functions (when the
       service supports it).

    Additional functionality that this class transparently offer:
            - discovery the capabilities of the service and their access
            methods
            - handling of transient errors with retrials
            - use of the appropriate credentials when interacting with the
            service
            - proper agent identification for logging on the server

    Three arguments are required in order to instantiate this class:
        1. A valid resource ID corresponding to the Web service that is
        being accessed. cadcutils.net.get_resources lists all the available
        resources at BOOTSTRAP_LOCATION location
        2. A cadcutils.net.Subject instance that contains the user credentials
        (no arguments for anonymous subject, certificate file or basic
        authentication).
        3. Agent is the name and version of the application as it will be
        presented in the HTTP
          request e.g. foo/1.0.2
    Once the BaseWsClient has been instantiated, the get/put/post/delete/head
    functions can be called on it.

    TODO: this implementation is very tightly coupled with the requests
    package, especially the arguments of get/post/put/delete/head that are
    blindly passed and the response object that is returned and the clients
    work with.
       """

    def __init__(self, resource_id, subject, agent, retry=True, host=None,
                 session_headers=None, insecure=False, idempotent_posts=False,
                 server_versions=None):
        """
        Client constructor
        :param resource_id -- ID of the resource being accessed (URI format)
        as it appears in the registry.
        :param subject -- The subject that is using the service
        :type subject: cadcutil.auth.Subject
        :param agent -- Name of the agent (application) that accesses the
        service and its version, e.g. foo/1.0.2
        :type agent: Subject
        :param retry -- True if the client retries on transient errors False
        otherwise
        :param host -- override the name of the host the service is running on
        (for testing purposes)
        :param session_headers -- Headers used throughout the session -
        dictionary format expected.
        :param insecure -- Allow insecure connections over SSL
        :param idempotent_post -- True if all HTTP POSTs can be considered
        idempotent, either because the server can deal with duplicate POSTs or
        because there's a higher level mechanism to deal with this. Idempotent
        POSTs can be automatically re-tried making them more fault tolerant.
        :param server_versions -- Dictionary of server names and corresponding
        supported versions (e.g. {'storage-inventory/raven': '1.0'}) that the
        client uses. This versions are checked for compatibility between cliet
        and server using the `server` HTTP header. If the version is behind
        the server version, a RuntimeException is raised prompting user to
        upgrade the software.
        """

        self.logger = logging.getLogger('BaseWsClient')
        logging.getLogger('BaseWsClient').addHandler(logging.NullHandler())

        if resource_id is None:
            raise ValueError('No resource ID provided')
        if agent is None or not agent:
            raise ValueError('agent is None or empty string')

        self._session = None
        self.subject = subject
        self.resource_id = resource_id
        self.retry = retry
        self.session_headers = session_headers
        self.verify = not insecure
        self.idempotent_posts = idempotent_posts
        self._server_versions = server_versions

        # agent is / delimited key value pairs, separated by a space,
        # containing the application name and version,
        # plus the name and version of application libraries.
        # eg: foo/1.0.2 foo-lib/1.2.3
        self.agent = agent

        # Get the package name and version, plus any imported libraries.
        self.package_info = "cadcutils/{} requests/{}".format(
            cadctools_version.version,
            requests.__version__)
        self.python_info = "{}/{}".format(platform.python_implementation(),
                                          platform.python_version())
        self.system_info = "{}/{}".format(platform.system(),
                                          platform.version())
        o_s = sys.platform
        if o_s.lower().startswith('linux'):
            distname = distro.name()
            version = distro.version()
            self.os_info = "{} {}".format(distname, version)
        elif o_s == "darwin":
            release, version, machine = platform.mac_ver()
            self.os_info = "Mac OS X {}".format(release)
        elif o_s.lower().startswith("win32"):
            release, version, csd, ptype = platform.win32_ver()
            self.os_info = "{} {}".format(release, version)

        # build the corresponding capabilities instance
        self.caps = WsCapabilities(self, host)
        self._host = host
        if resource_id.startswith('http'):
            self._host = resource_id
        if self._host is None:
            base_url = self.caps.get_access_url(SERVICE_AVAILABILITY_ID)
            self._host = urlparse(base_url).hostname

        # Clients should add entries to this dict for specialized
        # conversion of HTTP error codes into particular exceptions.
        #
        # Use this form to include a search string in the response to
        # handle multiple possibilities for a single HTTP code.
        #     XXX : {'SEARCHSTRING1' : exceptionInstance1,
        #            'SEARCHSTRING2' : exceptionInstance2}
        #
        # Otherwise provide a simple HTTP code -> exception mapping
        #     XXX : exceptionInstance
        #
        # The actual conversion is performed by get_exception()
        self._HTTP_STATUS_CODE_EXCEPTIONS = {
            401: exceptions.UnauthorizedException()}

    @property
    def host(self):
        return self._host

    def post(self, resource=None, **kwargs):
        """Wrapper for POST so that we use this client's session
           :param resource represents the resource to access. It can take two
           forms: 1 - a URL or 2 - a tuple representing a Web Service resource
           in which the first member of the tuple is the URI of the resource
           (capability) and the second argument is the path.
           :param kwargs additional arguments to pass to the requests.post
           :returns response as received from the request library
        """
        session = self._get_session()
        session.server_versions = self._server_versions
        return session.post(self._get_url(resource), verify=self.verify, **kwargs)

    def put(self, resource=None, **kwargs):
        """Wrapper for PUT so that we use this client's session
           :param resource represents the resource to access. It can take two
           forms: 1 - a URL or 2 - a tuple representing a Web Service resource
           in which the first member of the tuple is the URI of the resource
           (capability) and
           the second argument is the path.
           :param kwargs additional arguments to pass to the requests.post
           :returns response as received from the request library
        """
        session = self._get_session()
        session.server_versions = self._server_versions
        return session.put(self._get_url(resource), verify=self.verify, **kwargs)

    def get(self, resource, params=None, **kwargs):
        """Wrapper for GET so that we use this client's session
           :param resource represents the resource to access. It can take two
           forms: 1 - a URL or 2 - a tuple representing a Web Service resource
           in which the first member of the tuple is the URI of the resource
           (capability) and the second argument is the path.
           :param params parameters to use with get
           :param kwargs additional arguments to pass to the requests.post
           :returns response as received from the request library
        """
        session = self._get_session()
        session.server_versions = self._server_versions
        return session.get(self._get_url(resource), params=params, verify=self.verify, **kwargs)

    def delete(self, resource=None, **kwargs):
        """Wrapper for DELETE so that we use this client's session
           :param resource represents the resource to access. It can take two
           forms: 1 - a URL or 2 - a tuple representing a Web Service resource
           in which the first member of the tuple is the URI of the resource
           (capability) and the second argument is the path.
           :param kwargs additional arguments to pass to the requests.post
           :returns response as received from the request library
        """
        session = self._get_session()
        session.server_versions = self._server_versions
        return session.delete(self._get_url(resource), verify=self.verify, **kwargs)

    def head(self, resource=None, **kwargs):
        """Wrapper for HEAD so that we use this client's session
           :param resource represents the resource to access. It can take two
           forms: 1 - a URL or 2 - a tuple representing a Web Service resource
           in which the first member of the tuple is the URI of the resource
           (capability) and the second argument is the path.
           :param kwargs additional arguments to pass to the requests.post
           :returns response as received from the request library
        """
        session = self._get_session()
        session.server_versions = self._server_versions
        return session.head(self._get_url(resource), verify=self.verify, **kwargs)

    def is_available(self):
        """
        Checks whether the service is currently available or not
        :return: True if service is available, False otherwise
        """
        try:
            self.get((SERVICE_AVAILABILITY_ID, None))
        except exceptions.HttpException:
            return False
        return True

    def _get_url(self, resource):
        if type(resource) is tuple:
            # this is WS feature / path request
            path = ''
            if (resource[1] is not None) and (len(resource[1]) > 0):
                path = '/{}'.format(resource[1].strip('/'))
            if (len(resource) > 2):
                interface_type = resource[2]
                base_url = self.caps.get_access_url(resource[0],
                                                    interface_type)
            else:
                base_url = self.caps.get_access_url(resource[0])
            access_url = '{}{}'.format(base_url, path)
            return access_url
        else:
            # assume this is url.
            resource_url = urlparse(resource)
            if resource_url.scheme not in ['http', 'https']:
                raise ValueError('Incorrect resource URL: {}'.format(resource))
            return resource

    def _get_session(self):
        # Note that the cert goes into the adapter, but we can also
        # use name/password for the auth. We may want to enforce the
        # usage of only the cert in case both name/password and cert
        # are provided.
        if self._session is None:
            self.logger.debug('Creating session.')
            self._session = RetrySession(
                self.retry, idempotent_posts=self.idempotent_posts)
            # prevent requests from using .netrc
            self._session.trust_env = False
            if self.subject.certificate is not None:
                self._session.cert = (
                    self.subject.certificate, self.subject.certificate)
            elif self.subject.cookies:
                for cookie in self.subject.cookies:
                    cookie_obj = requests.cookies.create_cookie(
                        domain=cookie.domain, name=cookie.name,
                        value=cookie.value)
                    self._session.cookies.set_cookie(cookie_obj)
            else:
                if (not self.subject.anon) and (self.host is not None) and \
                        (self.subject.get_auth(self.host) is not None):
                    self._session.auth = self.subject.get_auth(self.host)

        user_agent = "{} {} {} {} ({})".format(self.agent, self.package_info,
                                               self.python_info,
                                               self.system_info, self.os_info)
        self._session.headers.update({"User-Agent": user_agent})
        if self.session_headers is not None:
            for header in self.session_headers:
                self._session.headers.update(self.session_headers)
        self._session.verify = self.verify
        return self._session


class BaseDataClient(BaseWsClient):
    """
    Base class for clients that interact with the CADC storage system (cadcdata
    and vos). Provides utilities for uploading and downloading files
    """

    def upload_file(self, url, src, md5_checksum=None, **kwargs):
        """Method to upload a file to CADC storage (archive or vospace). This
           method takes advantage of features in CADC services that uses
           PUTs with transactions in order to optimize and make the transfer
           more robust. A detailed description of the mechanism can be found
           at https://github.com/opencadc/storage-inventory/blob/master/minoc/PutTransaction.md
           :param url: URL to upload the file to
           :param src: name of the file to upload
           :param md5_checksum: optional md5 checksum of the file content. If
           available, the caller should set the attribute, otherwise the method
           might compute it (for small files) and introduce overhead.
           :param kwargs: other http attributes
           :throws: HttpExceptions
        """
        stat_info = os.stat(src)
        if stat_info.st_size == 0:
            raise ValueError('Cannot upload empty files')
        if HEADERS not in kwargs:
            kwargs[HEADERS] = {}
        orig_headers = kwargs.get(HEADERS)
        orig_headers[HTTP_LENGTH] = str(stat_info.st_size)

        def combine_headers(new_headers):
            # return a new dictionary that combines the original headers
            # and the new headers
            result = dict(orig_headers)
            result.update(new_headers)
            return result

        src_md5 = md5_checksum
        if not src_md5 and stat_info.st_size <= MAX_MD5_COMPUTE_SIZE:
            hash_md5 = hashlib.md5()
            with open(src, "rb") as f:
                for chunk in iter(lambda: f.read(4096), b""):
                    hash_md5.update(chunk)
            src_md5 = hash_md5.hexdigest()

        if src_md5:
            net.add_md5_header(headers=orig_headers, md5_checksum=src_md5)
            if stat_info.st_size < FILE_SEGMENT_THRESHOLD:
                # no transactions needed.
                retries = MD5_MISMATCH_RETRY
                while retries:
                    try:
                        with open(src, 'rb') as reader:
                            response = self._get_session().put(
                                url,
                                data=reader,
                                verify=self.verify, **kwargs)
                        self.logger.debug('{} uploaded (HTTP {})'.format(
                            src, response.status_code))
                        return
                    except exceptions.PreconditionFailedException as e:
                        # retry as this is likely caused by md5 mismatch
                        if not retries:
                            raise e
                        retries -= 1

        if stat_info.st_size < FILE_SEGMENT_THRESHOLD:
            # one go upload with transaction
            retries = MD5_MISMATCH_RETRY
            while retries:
                with util.Md5File(src, 'rb') as reader:
                    kwargs[HEADERS] = combine_headers(
                        {PUT_TXN_OP: PUT_TXN_START})
                    response = self._get_session().put(
                        url,
                        data=reader,
                        verify=self.verify, **kwargs)
                trans_id = response.headers[PUT_TXN_ID]
                # check the file made it OK
                dest_md5 = net.extract_md5(response.headers)
                if dest_md5 != reader.md5_checksum:
                    msg = 'File {} not properly uploaded. ' \
                          'Mismatched md5 src vs dest: {} vs {}'.format(
                            src, reader.md5_checksum, dest_md5)
                    self.logger.warning(msg)
                    # abort transaction
                    kwargs[HEADERS] = combine_headers(
                        {PUT_TXN_ID: trans_id,
                         PUT_TXN_OP: PUT_TXN_ABORT,
                         HTTP_LENGTH: '0'})
                    self._get_session().post(url, verify=self.verify, **kwargs)
                    retries -= 1
                    if retries:
                        self.logger.warning('Retrying')
                        continue
                    else:
                        raise exceptions.TransferException(msg)
                    # commit tran
                kwargs[HEADERS] = combine_headers({
                    PUT_TXN_ID: trans_id,
                    PUT_TXN_OP: PUT_TXN_COMMIT,
                    HTTP_LENGTH: '0'})
                self._get_session().put(url, verify=self.verify, **kwargs)
                return

        # large file that requires multiple segments
        kwargs[HEADERS] = combine_headers({
            HTTP_LENGTH: '0',
            PUT_TXN_TOTAL_LENGTH: str(stat_info.st_size),
            PUT_TXN_OP: PUT_TXN_START})
        response = self._get_session().put(url,
                                           verify=self.verify,
                                           **kwargs)
        trans_id = response.headers[PUT_TXN_ID]
        self.logger.debug('Starting transaction {} on url {}'.format(
            trans_id, url))
        min_segment = response.headers.get(PUT_TXN_MIN_SEGMENT, None)
        max_segment = response.headers.get(PUT_TXN_MAX_SEGMENT, None)
        seg_size = self._get_segment_size(stat_info.st_size,
                                          min_segment,
                                          max_segment)
        current_size = 0
        last_digest = hashlib.md5()
        try:
            # Obs -(-stat_info.st_size//seg_size) - ceiling division in PYTHON
            for segment in range(0, -(-stat_info.st_size//seg_size)):
                cur_seg_size = min(seg_size,
                                   stat_info.st_size-segment*seg_size)
                self.logger.debug('Sending segment {} of size {}'.format(
                    segment, cur_seg_size))
                # Note: setting the content length here is irrelevant as
                # requests is going to override it according to the size
                # of the data (as returned by Md5File file handler)
                current_size += cur_seg_size
                retries = MD5_MISMATCH_RETRY
                while retries:
                    try:
                        with util.Md5File(src, 'rb', segment*seg_size,
                                          cur_seg_size) as reader:
                            reader._md5_checksum = last_digest.copy()
                            kwargs[HEADERS] = combine_headers({
                                PUT_TXN_ID: trans_id,
                                HTTP_LENGTH: str(cur_seg_size)})
                            response = self._get_session().put(
                                url,
                                data=reader,
                                verify=self.verify,
                                **kwargs)
                    except exceptions.TransferException as e:
                        self.logger.warning('Errors transfering {} to {}: {}'.
                                            format(src, url, e.msg))
                        try:
                            kwargs[HEADERS] = combine_headers(
                                {PUT_TXN_ID: trans_id,
                                 HTTP_LENGTH: '0'})
                            response = self._get_session().head(
                                url,
                                **kwargs)
                        except Exception as e:
                            self.logger.error(
                                'Could not retrieve transaction {} '
                                'status from {}: {}'.format(trans_id, url,
                                                            str(e)))
                            raise e
                    # check the file made it OK
                    src_md5 = reader.md5_checksum
                    dest_md5 = net.extract_md5(response.headers)
                    if src_md5 != dest_md5:
                        msg = 'File {} not properly uploaded. ' \
                              'Mismatched md5 src vs dest: {} vs {}'.format(
                                src, src_md5, dest_md5)
                        self.logger.warning(msg)
                        retries -= 1
                        if retries:
                            # dest_md5 == None is the start state
                            if dest_md5 and \
                                    (dest_md5 != last_digest.hexdigest()):
                                self.logger.debug('Reverting transaction')
                                kwargs[HEADERS] = combine_headers(
                                    {PUT_TXN_ID: trans_id,
                                     PUT_TXN_OP: PUT_TXN_REVERT,
                                     HTTP_LENGTH: '0'})
                                response = self._get_session().post(
                                    url,
                                    verify=self.verify,
                                    **kwargs)
                                dest_md5 = net.extract_md5(response.headers)
                            if dest_md5 is None or \
                                    dest_md5 == last_digest.hexdigest():
                                self.logger.warning('Retrying')
                                continue
                            else:
                                self.logger.error(
                                    'BUG: reverted transaction does not match '
                                    'last md5: {} != {}'.format(
                                        dest_md5, last_digest.hexdigest()))
                        raise exceptions.TransferException(msg)
                    last_digest = reader._md5_checksum
                    break
        except BaseException as e:
            if trans_id:
                # abort transaction
                self.logger.debug('Aborting transaction {}'.format(trans_id))
                kwargs[HEADERS] = combine_headers({PUT_TXN_ID: trans_id,
                                                   PUT_TXN_OP: PUT_TXN_ABORT,
                                                   HTTP_LENGTH: '0'})
                self._get_session().post(url, verify=self.verify, **kwargs)
                self.logger.warning('Transaction {} aborted'.format(trans_id))
                raise e

        # commit tran
        self.logger.debug('Commit transaction')
        kwargs[HEADERS] = combine_headers({
            PUT_TXN_ID: trans_id,
            PUT_TXN_OP: PUT_TXN_COMMIT,
            HTTP_LENGTH: '0'})
        self._get_session().put(url, verify=self.verify, **kwargs)

    @staticmethod
    def _get_segment_size(file_size, min_segment, max_segment):
        segment_size = min(file_size, PREFERRED_SEGMENT_SIZE)
        if min_segment:
            segment_size = max(segment_size, int(min_segment))
        if max_segment:
            segment_size = min(segment_size, int(max_segment))
        return segment_size

    @staticmethod
    def _resolve_destination_file(dest, src_md5, default_file_name):
        # returns destination absolute file name as well as a temporary
        # destination to be used during the transfer
        if (dest is None) and (default_file_name is None):
            raise ValueError('BUG: Cannot resolve file name')
        if dest is not None:
            # got a dest name?
            if os.path.isdir(dest):
                final_dest = os.path.join(dest, default_file_name)
            else:
                final_dest = dest
        else:
            final_dest = os.path.basename(default_file_name)
        if src_md5:
            dir_name = os.path.dirname(final_dest)
            file_name = os.path.basename(final_dest)
            temp_dest = os.path.join(
                dir_name, '{}-{}.part'.format(file_name, src_md5))
            return final_dest, temp_dest
        else:
            return final_dest, final_dest

    def download_file(self, url, dest=None, **kwargs):
        """Method to download a file from CADC storage (archive or vospace).
           This method takes advantage of the HTTP Range feature available
           with the CADC services to optimize and make the transfer more
           robust.
           :param url: URL to get the file from
           :param dest: name of the file to store it to. If it's the name of
           the directory to save it to, it will use the Content-Disposition for
           the file name. By default, it saves the file in the current
           directory.
           :param kwargs: other http attributes
           :return: requests.Response object
           :throws: HttpExceptions

        """
        response = self.get(url, stream=True, **kwargs)
        src_md5 = net.extract_md5(response.headers)
        src_size = int(response.headers.get(HTTP_LENGTH, 0))
        content_disp = net.get_header_filename(response.headers)
        if hasattr(dest, 'read'):
            dest.write(response.raw.read())
            return
        else:
            final_dest, temp_dest = self._resolve_destination_file(
                dest=dest, src_md5=src_md5, default_file_name=content_disp)
            if os.path.isfile(final_dest) and \
               src_size == os.stat(final_dest).st_size and \
               src_md5 == \
                    hashlib.md5(open(final_dest, 'rb').read()).hexdigest():
                # nothing to be done
                return
            if src_md5 and src_size and os.path.isfile(temp_dest):
                stat_info = os.stat(temp_dest)
                if not stat_info.st_size or stat_info.st_size >= src_size:
                    # Note: the existence of a complete temporary file should
                    # be a bug. It's more likely that it's corrupted hence the
                    # removal below
                    os.remove(temp_dest)
                else:
                    if response.headers.get('Accept-Ranges', None) and \
                       response.headers.get('Accept-Ranges').strip() == \
                            'bytes':
                        # do a range request
                        response.raw.close()  # close existing stream
                        headers = {
                            'Range': 'bytes={}-'.format(stat_info.st_size)}
                        response = self.get(url, stream=True, headers=headers)
                        # at some point the warnings below should become errors
                        # even if the code can deal with a 200 response as well
                        # right now, these can be useful in debugging
                        if response.status_code != \
                                requests.codes.partial_content:
                            self.logger.warning(
                                'Expected partial content for range request')
                        exp_cr = 'bytes {}-{}/{}'.format(stat_info.st_size,
                                                         src_size - 1,
                                                         src_size)
                        actual_cr = response.headers.get('Content-Range', '')
                        if actual_cr != exp_cr:
                            self.logger.warning(
                                'Content-Range expected {} vs '
                                'received {}'.format(exp_cr, actual_cr))
            # need to send the original file content-length. The Range response
            # contains the content-length of the range.
            self._save_bytes(response, src_size, temp_dest, None)
            os.rename(temp_dest, final_dest)

    def _save_bytes(self, response, src_length, dest_file, process_bytes=None):
        # requests automatically decompresses the data.
        # Tell it to do it only if it had to
        hash_md5 = hashlib.md5()
        src_md5 = net.extract_md5(response.headers)

        class RawRange(object):
            """
            Wrapper class to make response.raw.read work as iterator and behave
            the same way as the corresponding response.iter_content. Useful
            with a progress bar.
            """

            def __init__(self, rsp):
                """
                :param rsp: HTTP response object
                """
                self._read = rsp.raw.read
                self.block_size = 0

            def __iter__(self):
                return self

            def __next__(self):
                return self.next()

            def next(self):
                # reads the next raw block
                data = self._read(self.block_size)
                if len(data) > 0:
                    if src_md5:
                        hash_md5.update(data)
                    return data
                else:
                    raise StopIteration()

            def get_instance(self, block_size):
                self.block_size = block_size
                return self

        update_mode = 'wb'
        dest_length = 0
        if os.path.isfile(dest_file) and os.stat(dest_file).st_size > 0:
            if response.status_code == requests.codes.partial_content:
                # Can resume download. Digest existing content on disk first
                update_mode = 'ab'
                dest_length = os.stat(dest_file).st_size
                with open(dest_file, 'rb') as f:
                    for chunk in iter(lambda: f.read(4096), b""):
                        hash_md5.update(chunk)
            else:
                os.remove(dest_file)

        rr = RawRange(response)
        reader = rr.get_instance
        # TODO - progress bar
        # if logger.isEnabledFor(logging.INFO) and total_length:
        #     chunks = progress.bar(reader(
        #         READ_BLOCK_SIZE),
        #         expected_size=((total_length / READ_BLOCK_SIZE) + 1))
        # else:
        dest_downloaded = 0
        chunks = reader(READ_BLOCK_SIZE)
        start = time.time()
        with open(dest_file, update_mode) as dest:
            for chunk in chunks:
                if process_bytes is not None:
                    process_bytes(chunk)
                dest.write(chunk)
                dest_downloaded += len(chunk)
        dest_md5 = hash_md5.hexdigest()
        dest_length += dest_downloaded
        if src_length and src_length != dest_length:
            error_msg = 'Sizes of source and downloaded file do not match: ' \
                        '{} vs {}'.format(src_length, dest_length)
        elif (src_md5 is not None) and (src_md5 != dest_md5):
            error_msg = 'Downloaded file is corrupted: expected md5({}) != ' \
                        'actual md5({})'.format(src_md5, dest_md5)
        else:
            duration = time.time() - start
            self.logger.info(
                'Successfully downloaded file {} in {}s '
                '(avg. speed: {}MB/s)'.format(
                    dest_file, round(duration, 2),
                    round(dest_downloaded / 1024 / 1024 / duration, 2)))
            return
        if not src_length or not src_md5:
            # clean up the temporary file
            os.remove(dest_file)
        raise exceptions.TransferException(error_msg)


def _check_server_version(supported_versions, server_header):
    if not supported_versions or not server_header:
        return
    for server in supported_versions.keys():
        logging.debug("Check server {}".format(server_header))
        if server in server_header:
            s_major, s_minor = [int(x) for x in supported_versions[server].split('.')]
            actual_version = server_header.split(server + '-')
            if len(actual_version) != 2:
                raise ValueError('Cannot parse server version: ' + server_header)
            # just in case more info is added after version
            actual_version = actual_version[1].split(' ')[0]
            if actual_version.count('.') < 1:
                raise ValueError('Expected at least major.minor server version: ' + server_header)
            a_major, a_minor = [int(x) for x in actual_version.split('.')[:2]]
            if a_major <= s_major and a_minor <= s_minor:
                return
            logging.debug(
                'Version incompatibility for {} - Client({}.{}) vs Server({}.{})'.format(
                    server, s_major, s_minor, a_major, a_minor))
            raise RuntimeError(
                'Client and server software not compatible anymore. Please upgrade application.')


class RetrySession(Session):
    """ Session that automatically does a number of retries for failed
        transient errors. The time between retries double every time until a
        maximum of 30sec is reached

        This does not work with POST method

        The following network errors are considered transient:
            requests.codes.unavailable,
            requests.codes.service_unavailable,
            requests.codes.gateway_timeout,
            requests.codes.request_timeout,
            requests.codes.timeout,
            requests.codes.precondition_failed,
            requests.codes.precondition,
            requests.codes.payment_required,
            requests.codes.payment

        In addition, the Connection error 'Connection reset by remote user'
        also triggers a retry
        """

    retry_errors = [requests.codes.unavailable,
                    requests.codes.service_unavailable,
                    requests.codes.gateway_timeout,
                    requests.codes.request_timeout,
                    requests.codes.timeout,
                    requests.codes.payment_required,
                    requests.codes.payment]

    def __init__(self, retry=True, start_delay=1, idempotent_posts=False,
                 *args, **kwargs):
        """
        ::param retry: set to False if retries not required
        ::param start_delay: start delay interval between retries (default=1s).
                Note that for HTTP 503, this code follows the retry timeout
                set by the server in Retry-After
        ::param idempotent_posts: POST requests in general are not idempotent
        and they are not automatically re-tried on failures. Setting this flag
        to true can override that, in case when a specific client-server
        implementation can handle duplicate POST requests at a higher level.
        """
        self.logger = logging.getLogger('RetrySession')
        self.retry = retry
        self.start_delay = start_delay
        self.idempotent_posts = idempotent_posts
        super(RetrySession, self).__init__(*args, **kwargs)
        self._server_versions = None  # server versions that client supports

    @property
    def server_versions(self):
        return self._server_versions

    @server_versions.setter
    def server_versions(self, versions):
        self._server_versions = versions

    def send(self, request, **kwargs):
        """
        Send a given PreparedRequest, wrapping the connection to service in
        try/except that retries on
        Connection reset by peer.
        :param request: The prepared request to send._session
        :param kwargs: Any keywords the adaptor for the request accepts.
        :return: the response
        :rtype: requests.Response
        """
        # merge kwargs with env
        proxies = kwargs.get('proxies') or {}
        settings = self.merge_environment_settings(
            request.url, proxies, kwargs.get('stream'), kwargs.get('verify'),
            kwargs.get('cert'))
        kwargs.update(settings)

        # requests does not provide a default timeout, hence we might need
        # to add it
        if 'timeout' not in kwargs or kwargs['timeout'] is None:
            kwargs['timeout'] = 120

        if (request.method.upper() == 'POST') and self.idempotent_posts:
            self.logger.debug(
                'POST requests considered idempotent. re-tries enabled')

        if (request.method.upper() != 'POST' or self.idempotent_posts) \
           and self.retry:
            current_delay = max(self.start_delay, DEFAULT_RETRY_DELAY)
            current_delay = min(current_delay, MAX_RETRY_DELAY)
            num_retries = 0
            self.logger.debug(
                "Sending request {0}  to server.".format(request))
            current_error = None
            while True:
                try:
                    response = super(RetrySession, self).send(request,
                                                              **kwargs)
                    self.check_status(response)
                    return response
                except requests.exceptions.ConnectTimeout as ct:
                    # retry on timeouts
                    current_error = ct
                    self.logger.debug(ct)
                except requests.exceptions.ReadTimeout as rt:
                    # this could happen after the request has made it to
                    # the server so it should be re-done
                    raise exceptions.TransferException(
                        'Read timeout on {}'.format(request.url), rt)
                except requests.HTTPError as e:
                    if e.response.status_code not in self.retry_errors:
                        raise exceptions.HttpException(e)
                    current_error = e
                    if e.response.status_code == requests.codes.unavailable:
                        # is there a delay from the server (Retry-After)?
                        try:
                            current_delay = int(
                                e.response.headers.get(SERVICE_RETRY,
                                                       current_delay))
                            current_delay = min(current_delay, MAX_RETRY_DELAY)
                        except Exception:
                            pass
                except requests.ConnectionError as ce:
                    if 'Connection reset by peer' in str(ce):
                        # Likely a network error that the caller can re-try
                        raise exceptions.TransferException(
                            'Transfer error on URL: {}'.format(request.url))
                    else:
                        # Can't recover (bad url, etc)
                        raise exceptions.HttpException(orig_exception=ce)
                if num_retries == MAX_NUM_RETRIES:
                    break
                self.logger.debug(
                    "Error {}. Resending request in {}s ...".format(
                        str(current_error), current_delay))
                time.sleep(current_delay)
                num_retries += 1
                current_delay = min(current_delay * 2, MAX_RETRY_DELAY)
            raise exceptions.HttpException(current_error)
        else:
            response = super(RetrySession, self).send(request, **kwargs)
            self.check_status(response, retry=False)
            return response

    def check_status(self, response, retry=True):
        """
        Check the response status. Maps the application related requests
        error status into Exceptions and raises the others
        :param response: response
        :param retry: request can be re-tried. Let the re-tried errors through
        :return:
        """
        _check_server_version(self.server_versions, response.headers.get('server', None))
        try:
            response.raise_for_status()
        except requests.HTTPError as e:
            if e.response.status_code == requests.codes.not_found:
                raise exceptions.NotFoundException(orig_exception=e)
            elif e.response.status_code == requests.codes.unauthorized:
                raise exceptions.UnauthorizedException(orig_exception=e)
            elif e.response.status_code == requests.codes.forbidden:
                raise exceptions.ForbiddenException(orig_exception=e)
            elif e.response.status_code == requests.codes.bad_request:
                raise exceptions.BadRequestException(orig_exception=e)
            elif e.response.status_code == requests.codes.precondition_failed:
                raise exceptions.PreconditionFailedException(orig_exception=e)
            elif e.response.status_code == requests.codes.conflict:
                raise exceptions.AlreadyExistsException(orig_exception=e)
            elif e.response.status_code == \
                    requests.codes.internal_server_error:
                raise exceptions.InternalServerException(orig_exception=e)
            elif e.response.status_code == \
                    requests.codes.request_entity_too_large:
                raise exceptions.ByteLimitException(orig_exception=e)
            elif retry and e.response.status_code in self.retry_errors:
                raise e
            else:
                raise exceptions.UnexpectedException(orig_exception=e)


DEFAULT_REGISTRY = \
    'https://ws.cadc-ccda.hia-iha.nrc-cnrc.gc.ca/reg/resource-caps'
REG_REFRESH_INTERVAL = 10 * 60
CACHE_LOCATION = os.path.join(os.path.expanduser("~"), '.config',
                              'cadc-registry')
REGISTRY_FILE = 'resource-caps'


class WsCapabilities(object):
    """
    Contains the capabilities of Web Services. The most useful function is
    get_access_url that returns the url corresponding to a feature of a
    Web Service
    """

    def __init__(self, ws_client, host=None):
        """
        :param ws_client: WebService client that the capabilities are required
        for
        :param host: use this host rather than the default CADC host for
        reg lookup
        """
        self.logger = logging.getLogger('WsCapabilities')
        self.ws = ws_client
        self._host = host
        cache_location = CACHE_LOCATION
        if host is not None:
            cache_location = os.path.join(cache_location, 'alt-domains', host)

        if not os.path.isdir(cache_location):
            os.makedirs(cache_location)

        # check the registry file in cache if it requires a refresh
        self.reg_file = os.path.join(cache_location, REGISTRY_FILE)
        resource_id = urlparse(ws_client.resource_id)
        self.caps_file = os.path.join(cache_location, resource_id.netloc,
                                      resource_id.path.strip('/'))
        # the name of the cached capabilities files is preceeded by a ".". Old
        # files don't have this rule so delete them if they exist. The
        # following block should be eventually removed
        if os.path.isfile(self.caps_file):
            os.remove(self.caps_file)
        # prefix the name of the file with '.' to avoid collisions with
        # subdirectory names
        self.caps_file = '/.'.join(self.caps_file.rsplit('/', 1))
        self.last_regtime = 0
        self.last_capstime = 0
        self._caps_reader = wscapabilities.CapabilitiesReader()
        self.caps_urls = {}
        self.features = {}
        self.capabilities = {}

    def get_access_url(self, feature, interface_type='vs:ParamHTTP'):
        """
        Returns the access URL corresponding to a feature and the
        authentication information associated with the subject that created
        the Web Service client
        :param feature: Web Service feature
        :param interface_type: the type of the interface
        :return: corresponding access URL
        """

        if (time.time() - self.last_capstime) > REG_REFRESH_INTERVAL:
            caps = util.get_url_content(url=self._get_capability_url(),
                                        cache_file=self.caps_file,
                                        refresh_interval=REG_REFRESH_INTERVAL,
                                        verify=self.ws.verify)
            # caps is a string but it's xml content claims it's utf-8 encode,
            # hence need to encode it before
            # parsing it.
            self.capabilities = self._caps_reader.parsexml(
                caps.encode('utf-8'))
            if (time.time() - self.last_capstime) > REG_REFRESH_INTERVAL:
                self.last_capstime = time.time()
        sms = self.ws.subject.get_security_methods()

        return self.capabilities.get_access_url(feature, sms, interface_type)

    @property
    def host(self):
        return self._host

    def _get_capability_url(self):
        """
        Parses the registry information and returns the url of the
        capabilities feature of the Web Service
        :return: URL to the capabilities feature
        """
        if self.ws.resource_id.startswith('http'):
            return '{}/capabilities'.format(self.ws.resource_id)
        if (time.time() - self.last_regtime) > REG_REFRESH_INTERVAL:
            if self.last_regtime == 0:
                # startup
                try:
                    self.last_regtime = os.path.getmtime(self.reg_file)
                except OSError:
                    # cannot read the cache file for whatever reason
                    pass

            # replace registry host name if necessary
            registry_url = DEFAULT_REGISTRY
            url = urlparse(registry_url)
            if self._host is not None and url.netloc != self._host:
                registry_url = '{}://{}{}'.format(url.scheme, self._host,
                                                  url.path)
            self.logger.debug('Resolved URL: {}'.format(registry_url))

            reg = util.get_url_content(url=registry_url,
                                       cache_file=self.reg_file,
                                       refresh_interval=REG_REFRESH_INTERVAL,
                                       verify=self.ws.verify)
            self.caps_urls = {}
            if (time.time() - self.last_regtime) > REG_REFRESH_INTERVAL:
                self.last_regtime = time.time()
            # parse it
            for line in reg.split('\n'):
                if not line.startswith('#') and (len(line) > 0):
                    feature, url = line.split('=')
                    self.caps_urls[feature.strip()] = url.strip()
        if self.ws.resource_id not in self.caps_urls:
            raise AttributeError(
                'Resource ID {} not found. Available resource IDs: {}'.
                format(self.ws.resource_id, self.caps_urls.keys()))
        return self.caps_urls[self.ws.resource_id]



# -*- coding: utf-8 -*-
# ***********************************************************************
# ******************  CANADIAN ASTRONOMY DATA CENTRE  *******************
# *************  CENTRE CANADIEN DE DONNÉES ASTRONOMIQUES  **************
#
#  (c) 2022.                            (c) 2022.
#  Government of Canada                 Gouvernement du Canada
#  National Research Council            Conseil national de recherches
#  Ottawa, Canada, K1A 0R6              Ottawa, Canada, K1A 0R6
#  All rights reserved                  Tous droits réservés
#
#  NRC disclaims any warranties,        Le CNRC dénie toute garantie
#  expressed, implied, or               énoncée, implicite ou légale,
#  statutory, of any kind with          de quelque nature que ce
#  respect to the software,             soit, concernant le logiciel,
#  including without limitation         y compris sans restriction
#  any warranty of merchantability      toute garantie de valeur
#  or fitness for a particular          marchande ou de pertinence
#  purpose. NRC shall not be            pour un usage particulier.
#  liable in any event for any          Le CNRC ne pourra en aucun cas
#  damages, whether direct or           être tenu responsable de tout
#  indirect, special or general,        dommage, direct ou indirect,
#  consequential or incidental,         particulier ou général,
#  arising from the use of the          accessoire ou fortuit, résultant
#  software.  Neither the name          de l'utilisation du logiciel. Ni
#  of the National Research             le nom du Conseil National de
#  Council of Canada nor the            Recherches du Canada ni les noms
#  names of its contributors may        de ses  participants ne peuvent
#  be used to endorse or promote        être utilisés pour approuver ou
#  products derived from this           promouvoir les produits dérivés
#  software without specific prior      de ce logiciel sans autorisation
#  written permission.                  préalable et particulière
#                                       par écrit.
#
#  This file is part of the             Ce fichier fait partie du projet
#  OpenCADC project.                    OpenCADC.
#
#  OpenCADC is free software:           OpenCADC est un logiciel libre ;
#  you can redistribute it and/or       vous pouvez le redistribuer ou le
#  modify it under the terms of         modifier suivant les termes de
#  the GNU Affero General Public        la “GNU Affero General Public
#  License as published by the          License” telle que publiée
#  Free Software Foundation,            par la Free Software Foundation
#  either version 3 of the              : soit la version 3 de cette
#  License, or (at your option)         licence, soit (à votre gré)
#  any later version.                   toute version ultérieure.
#
#  OpenCADC is distributed in the       OpenCADC est distribué
#  hope that it will be useful,         dans l’espoir qu’il vous
#  but WITHOUT ANY WARRANTY;            sera utile, mais SANS AUCUNE
#  without even the implied             GARANTIE : sans même la garantie
#  warranty of MERCHANTABILITY          implicite de COMMERCIALISABILITÉ
#  or FITNESS FOR A PARTICULAR          ni d’ADÉQUATION À UN OBJECTIF
#  PURPOSE.  See the GNU Affero         PARTICULIER. Consultez la Licence
#  General Public License for           Générale Publique GNU Affero
#  more details.                        pour plus de détails.
#
#  You should have received             Vous devriez avoir reçu une
#  a copy of the GNU Affero             copie de la Licence Générale
#  General Public License along         Publique GNU Affero avec
#  with OpenCADC.  If not, see          OpenCADC ; si ce n’est
#  <http://www.gnu.org/licenses/>.      pas le cas, consultez :
#                                       <http://www.gnu.org/licenses/>.
#
#
# ***********************************************************************

import errno
import logging
import os
import sys
import configparser
from shutil import copyfile


logger = logging.getLogger('config')
logger.setLevel(logging.INFO)

if sys.version_info[1] > 6:
    logger.addHandler(logging.NullHandler())


def mkdir_p(path):
    try:
        os.makedirs(path)
    except OSError as exc:
        if exc.errno == errno.EEXIST and os.path.isdir(path):
            pass
        else:
            raise


class Config(object):

    def __init__(self, config_path, default_config_path=None):
        logger.info("Using config file {0}.".format(config_path))

        # check config file exists and can be read
        if not os.path.isfile(config_path) and not os.access(config_path,
                                                             os.R_OK):
            error = "Can not read {0}.".format(config_path)
            logger.debug(error)
            raise IOError(error)

        self.parser = configparser.ConfigParser()
        if default_config_path:
            try:
                self.parser.read_file(open(default_config_path))
            except configparser.Error as exc:
                logger.debug("Error opening {0} because {1}.".format(
                    default_config_path, exc.message))

        try:
            self.parser.read(config_path)
        except configparser.Error as exc:
            logger.debug("Error opening {0} because {1}.".format(
                config_path, exc.message))

    def get(self, section, option):
        try:
            return self.parser.get(section, option)
        except (configparser.NoOptionError, configparser.NoSectionError):
            pass
        return None

    @staticmethod
    def write_config(config_path, default_config_path):
        """
        :param config_path:
        :param default_config_path:
        :return:
        """

        # if not local config file then write the default config file
        if not os.path.isfile(config_path):
            mkdir_p(os.path.dirname(config_path))
            copyfile(default_config_path, config_path)
            return

        # read local config file
        parser = configparser.ConfigParser()
        try:
            parser.read(config_path)
        except configparser.Error as exc:
            logger.debug("Error opening {0} because {1}.".format(config_path,
                                                                 exc.message))
            return

        # read default config file
        default_parser = configparser.RawConfigParser()
        try:
            default_parser.read(default_config_path)
        except configparser.Error as exc:
            logger.debug("Error opening {0} because {1}.".format(
                default_config_path, exc.message))
            return

        # update config file with new options from the default config
        updated = False
        for section in default_parser.sections():
            default_items = default_parser.items(section)
            for option, value in default_items:
                if not parser.has_section(section):
                    parser.add_section(section)
                if not parser.has_option(section, option):
                    parser.set(section, option, value)
                    updated = True

        # remove old options not in the default config file?
        # for section in default_parser.sections():
        #     options = parser.options(section)
        #     for option in options:
        #         if not default_parser.has_option(section, option):
        #             parser.remove_option(section, option)

        # write updated config file
        if updated:
            try:
                config_file = open(config_path, 'w')
                parser.write(config_file)
                config_file.close()
            except Exception as exc:
                print(exc.message)



# -*- coding: utf-8 -*-

# ***********************************************************************
# ******************  CANADIAN ASTRONOMY DATA CENTRE  *******************
# *************  CENTRE CANADIEN DE DONNÉES ASTRONOMIQUES  **************
#
#  (c) 2022.                            (c) 2022.
#  Government of Canada                 Gouvernement du Canada
#  National Research Council            Conseil national de recherches
#  Ottawa, Canada, K1A 0R6              Ottawa, Canada, K1A 0R6
#  All rights reserved                  Tous droits réservés
#
#  NRC disclaims any warranties,        Le CNRC dénie toute garantie
#  expressed, implied, or               énoncée, implicite ou légale,
#  statutory, of any kind with          de quelque nature que ce
#  respect to the software,             soit, concernant le logiciel,
#  including without limitation         y compris sans restriction
#  any warranty of merchantability      toute garantie de valeur
#  or fitness for a particular          marchande ou de pertinence
#  purpose. NRC shall not be            pour un usage particulier.
#  liable in any event for any          Le CNRC ne pourra en aucun cas
#  damages, whether direct or           être tenu responsable de tout
#  indirect, special or general,        dommage, direct ou indirect,
#  consequential or incidental,         particulier ou général,
#  arising from the use of the          accessoire ou fortuit, résultant
#  software.  Neither the name          de l'utilisation du logiciel. Ni
#  of the National Research             le nom du Conseil National de
#  Council of Canada nor the            Recherches du Canada ni les noms
#  names of its contributors may        de ses  participants ne peuvent
#  be used to endorse or promote        être utilisés pour approuver ou
#  products derived from this           promouvoir les produits dérivés
#  software without specific prior      de ce logiciel sans autorisation
#  written permission.                  préalable et particulière
#                                       par écrit.
#
#  This file is part of the             Ce fichier fait partie du projet
#  OpenCADC project.                    OpenCADC.
#
#  OpenCADC is free software:           OpenCADC est un logiciel libre ;
#  you can redistribute it and/or       vous pouvez le redistribuer ou le
#  modify it under the terms of         modifier suivant les termes de
#  the GNU Affero General Public        la “GNU Affero General Public
#  License as published by the          License” telle que publiée
#  Free Software Foundation,            par la Free Software Foundation
#  either version 3 of the              : soit la version 3 de cette
#  License, or (at your option)         licence, soit (à votre gré)
#  any later version.                   toute version ultérieure.
#
#  OpenCADC is distributed in the       OpenCADC est distribué
#  hope that it will be useful,         dans l’espoir qu’il vous
#  but WITHOUT ANY WARRANTY;            sera utile, mais SANS AUCUNE
#  without even the implied             GARANTIE : sans même la garantie
#  warranty of MERCHANTABILITY          implicite de COMMERCIALISABILITÉ
#  or FITNESS FOR A PARTICULAR          ni d’ADÉQUATION À UN OBJECTIF
#  PURPOSE.  See the GNU Affero         PARTICULIER. Consultez la Licence
#  General Public License for           Générale Publique GNU Affero
#  more details.                        pour plus de détails.
#
#  You should have received             Vous devriez avoir reçu une
#  a copy of the GNU Affero             copie de la Licence Générale
#  General Public License along         Publique GNU Affero avec
#  with OpenCADC.  If not, see          OpenCADC ; si ce n’est
#  <http://www.gnu.org/licenses/>.      pas le cas, consultez :
#                                       <http://www.gnu.org/licenses/>.
#
#
# ***********************************************************************

"""
Web Service capabilities information
"""

import logging
from lxml import etree

from urllib.parse import urlparse

# nothing for public consumption
__all__ = []

logger = logging.getLogger(__name__)


class Capabilities(object):
    """
    Holds information regarding the capabilities of a Web Service.
    """

    def __init__(self):
        self._caps = {}

    def add_capability(self, capability):
        self._caps[capability.standard_id] = capability

    def get_access_url(self, capability_id, security_methods,
                       interface_type='vs:ParamHTTP'):
        """
        Returns the access URL corresponding to a capability ID and a list of
        security methods. Raises a ValueError if no entry found.
        :param capability_id: ID of the capability
        :param security_methods: lists of IDs of the security methods in the
        preferred order
        :param interface_type: type of the interface
        :return: URL to use for accessing the feature/capability
        """
        capability = self._caps[capability_id]
        sec_methods = security_methods if security_methods else []
        if capability is None:
            raise ValueError('Capability {} not found'.format(capability_id))
        # if the capability supports anonymous access and this is the only
        # capability interface or
        # the client comes with no security_methods, then return the url
        # corresponding to the anonymous access
        for sm in sec_methods:
            interface = capability.get_interface(sm, interface_type)
            if interface is not None:
                return interface.access_url
        if ((capability.get_interface(None) is not None) or
                (len(sec_methods) == 0)):
            i = capability.get_interface(None, interface_type)
            if i is not None:
                return i.access_url
            else:
                raise ValueError(
                    'Capability {} does not support annonymous access. '
                    'Please authenticate first'.
                    format(capability_id))
        raise ValueError('Capability {} does not support security methods {}'.
                         format(capability_id, security_methods))


def check_valid_url(url_str):
    """
    Checks whether the url argument as at least 3 components: scheme,
    netloc and path. Anything missing?
    :param url: url to check
    :raises ValueError if not valid
    """
    if url_str is None:
        raise ValueError('Invalid URL: {}'.format(url_str))
    url = urlparse(url_str)
    if (len(url.scheme) == 0) or \
            (len(url.netloc) == 0) or \
            (len(url.path) == 0):
        raise ValueError('Invalid URL: {}'.format(url_str))


class Capability(object):
    """
    Represents a capability of a Web Service. It has one or multiple
    _interfaces (access URLs)
    corresponding to different supported security methods
    """

    def __init__(self, standard_id):
        """
        :param standard_id: ID of the capability
        """
        # validate standard id is valid uri with 3 components: scheme,
        # netloc and path. Anything missing?
        check_valid_url(standard_id)
        self.standard_id = standard_id
        self._interfaces = []

    def get_interface(self, security_method, interface_type='vs:ParamHTTP'):
        # validate security_method is uri
        if security_method:
            check_valid_url(security_method)
        for i in self._interfaces:
            if (i.security_method == security_method):
                return i
        return None

    def add_interface(self, access_url, security_method,
                      interface_type='vs:ParamHTTP'):
        # validate arguments
        check_valid_url(access_url)
        if security_method is not None:
            check_valid_url(security_method)
        interface = self.get_interface(security_method, interface_type)
        if interface:
            # there's already an access url for this security method
            # HTTPS access urls are preferred, so keep that one that has
            # the https scheme
            old_url = urlparse(interface.access_url)
            new_url = urlparse(access_url)
            if (old_url.scheme == 'https') or (new_url.scheme == 'http'):
                logger.debug('access url already exists for {}'.
                             format(security_method))
                return
            else:
                interface.access_url = access_url
        else:
            self._interfaces.append(
                Interface(interface_type, security_method, access_url))

    @property
    def num_interfaces(self):
        """
        Number of supported interfaces
        :return: number of supported interfaces
        """
        return len(self._interfaces)


class Interface(object):
    """
    Class representing the interface of a capability

    """

    def __init__(self, type, security_method, access_url):
        self.type = type
        if security_method:
            check_valid_url(security_method)
        self.security_method = security_method
        self.access_url = access_url


class CapabilitiesReader(object):
    """
    Class to parse the capabilities xml file and return the corresponding
    capabilities object
    """

    def __init__(self):
        """
        Arguments:
        :param validate : If True enable schema validation, False otherwise
        """

    def parsexml(self, content):
        """
        Parses the content and returns the corresponding Capabilities object
        :param content: string, containing the capabilities xml document
        :return: corresponding Capabilities
        """
        caps = Capabilities()
        if not content:
            raise ValueError(
                'Cannot access remote service info (capabilities). Likely '
                'due to network error. Please re-try.')
        try:
            doc = etree.fromstring(content)
        except Exception as e:
            raise ValueError(
                'Cannot access service information (capabilities).', e)
        for cap in doc.iterfind('capability'):
            try:
                capability = Capability(cap.get('standardID'))
            except Exception:
                raise ValueError('Cannot read service info (capabilities). '
                                 'Capability standard ID is invalid URL: {}'.
                                 format(cap.get('standardID')))

            for child in cap.iterchildren(tag='interface'):
                interface_type = child.get('{{{}}}type'.
                                           format(doc.nsmap['xsi']))
                if (child.find('accessURL') is not None) \
                        and (child.find('accessURL').text is not None):
                    access_url = child.find('accessURL').text.strip()
                else:
                    raise ValueError('Cannot read service info (capabilities).'
                                     ' No accessURL for {}'.
                                     format(capability.standard_id))
                for sm in child.iterchildren('securityMethod'):
                    if sm.get('standardID') is not None:
                        security_method = \
                            sm.get('standardID').strip()
                    else:
                        security_method = None
                    capability.add_interface(access_url,
                                             security_method,
                                             interface_type)
                if capability.num_interfaces == 0:
                    # add default anonymous capability
                    capability.add_interface(access_url, None, interface_type)

            if len(capability._interfaces) == 0:
                raise ValueError('BUG reading service info (capabilities). '
                                 'No interfaces found for capability {}'.
                                 format(capability.standard_id))
            caps.add_capability(capability)
        if len(caps._caps) == 0:
            raise ValueError('BUG reading remote service info '
                             '(capabilities) - No actual capabilities found')
        return caps



# -*- coding: utf-8 -*-
# ***********************************************************************
# ******************  CANADIAN ASTRONOMY DATA CENTRE  *******************
# *************  CENTRE CANADIEN DE DONNÉES ASTRONOMIQUES  **************
#
#  (c) 2022.                            (c) 2022.
#  Government of Canada                 Gouvernement du Canada
#  National Research Council            Conseil national de recherches
#  Ottawa, Canada, K1A 0R6              Ottawa, Canada, K1A 0R6
#  All rights reserved                  Tous droits réservés
#
#  NRC disclaims any warranties,        Le CNRC dénie toute garantie
#  expressed, implied, or               énoncée, implicite ou légale,
#  statutory, of any kind with          de quelque nature que ce
#  respect to the software,             soit, concernant le logiciel,
#  including without limitation         y compris sans restriction
#  any warranty of merchantability      toute garantie de valeur
#  or fitness for a particular          marchande ou de pertinence
#  purpose. NRC shall not be            pour un usage particulier.
#  liable in any event for any          Le CNRC ne pourra en aucun cas
#  damages, whether direct or           être tenu responsable de tout
#  indirect, special or general,        dommage, direct ou indirect,
#  consequential or incidental,         particulier ou général,
#  arising from the use of the          accessoire ou fortuit, résultant
#  software.  Neither the name          de l'utilisation du logiciel. Ni
#  of the National Research             le nom du Conseil National de
#  Council of Canada nor the            Recherches du Canada ni les noms
#  names of its contributors may        de ses  participants ne peuvent
#  be used to endorse or promote        être utilisés pour approuver ou
#  products derived from this           promouvoir les produits dérivés
#  software without specific prior      de ce logiciel sans autorisation
#  written permission.                  préalable et particulière
#                                       par écrit.
#
#  This file is part of the             Ce fichier fait partie du projet
#  OpenCADC project.                    OpenCADC.
#
#  OpenCADC is free software:           OpenCADC est un logiciel libre ;
#  you can redistribute it and/or       vous pouvez le redistribuer ou le
#  modify it under the terms of         modifier suivant les termes de
#  the GNU Affero General Public        la “GNU Affero General Public
#  License as published by the          License” telle que publiée
#  Free Software Foundation,            par la Free Software Foundation
#  either version 3 of the              : soit la version 3 de cette
#  License, or (at your option)         licence, soit (à votre gré)
#  any later version.                   toute version ultérieure.
#
#  OpenCADC is distributed in the       OpenCADC est distribué
#  hope that it will be useful,         dans l’espoir qu’il vous
#  but WITHOUT ANY WARRANTY;            sera utile, mais SANS AUCUNE
#  without even the implied             GARANTIE : sans même la garantie
#  warranty of MERCHANTABILITY          implicite de COMMERCIALISABILITÉ
#  or FITNESS FOR A PARTICULAR          ni d’ADÉQUATION À UN OBJECTIF
#  PURPOSE.  See the GNU Affero         PARTICULIER. Consultez la Licence
#  General Public License for           Générale Publique GNU Affero
#  more details.                        pour plus de détails.
#
#  You should have received             Vous devriez avoir reçu une
#  a copy of the GNU Affero             copie de la Licence Générale
#  General Public License along         Publique GNU Affero avec
#  with OpenCADC.  If not, see          OpenCADC ; si ce n’est
#  <http://www.gnu.org/licenses/>.      pas le cas, consultez :
#                                       <http://www.gnu.org/licenses/>.
#
#
# ***********************************************************************
import logging
import sys
import inspect
from argparse import ArgumentParser, RawDescriptionHelpFormatter, SUPPRESS, \
    Action
from datetime import datetime
from urllib.parse import urlparse
from operator import attrgetter
import hashlib
import os
from packaging.version import Version
import requests
import time
import json
import warnings
from pathlib import Path
from cadcutils import exceptions

__all__ = ['IVOA_DATE_FORMAT', 'date2ivoa', 'str2ivoa', 'get_url_content',
           'get_logger', 'get_log_level', 'get_base_parser', 'Md5File',
           'check_version', 'VersionWarning']

# TODO both these are very bad, implement more sensibly
IVOA_DATE_FORMAT = "%Y-%m-%dT%H:%M:%S.%f"

CADC_CACHE_DIR = os.path.join(os.path.expanduser("~"), '.config')
VERSION_REFRESH_INTERVAL = 24 * 60 * 60  # 24h

DEFAULT_LOG_FORMAT = "%(levelname)s: %(name)s %(message)s"
DEBUG_LOG_FORMAT = "%(levelname)s: @(%(asctime)s) %(name)s " \
                   "%(module)s.%(funcName)s:%(lineno)d - %(message)s"

logger = logging.getLogger(__name__)


def date2ivoa(d):
    """
    Takes a datetime and returns a string formatted
    to the IVOA date format yyyy-MM-dd'T'HH:mm:ss.SSS
    """

    if d is None:
        return None
    return d.strftime(IVOA_DATE_FORMAT)[:23]


def str2ivoa(s):
    """
    Takes a IVOA date formatted string and returns a datetime.
    """

    if s is None:
        return None
    return datetime.strptime(s, IVOA_DATE_FORMAT)


def get_logger(namespace=None, log_level=logging.ERROR):
    """
    Create a logger with a standard format.
    :param namespace: The namespace to which to attach the logger.
        default: the name of the calling module.
    :ptype log_level: The initial log level.
        default: logging.ERROR
    :param log_level: logging level
    :return the logger object.
    """

    if namespace is None:
        frm = inspect.stack()[1]
        mod = inspect.getmodule(frm[0])
        if mod is None:
            namespace = __name__
        else:
            namespace = mod.__name__

    logger = logging.getLogger(namespace)

    if not len(logger.handlers):
        handler = logging.StreamHandler(sys.stdout)
        logger.addHandler(handler)
    else:
        handler = logger.handlers[0]

    log_format = DEFAULT_LOG_FORMAT
    if log_level == logging.DEBUG:
        log_format = DEBUG_LOG_FORMAT

    logger.setLevel(log_level)
    handler.setFormatter(logging.Formatter(fmt=log_format))

    return logger


def get_log_level(args):
    """
    Obtain a single logger level from parsed args.

    :param args The list of arguments.

    """

    log_level = ((args.debug and logging.DEBUG) or
                 (args.verbose and logging.INFO) or
                 (args.quiet and logging.FATAL) or
                 logging.ERROR)
    return log_level


def parse_resource_id(resource_id):
    """
    Parses a resource identifier and returns its components as a tuple:
    (scheme, netloc, path, params, query, fragment).
    :param resource_id: the resource identifier
    :return: (scheme, netloc, path, params, query, fragment)
    """
    result = urlparse(resource_id)
    if len(result.netloc) < 2 or len(result.path) < 2 or \
            (result.scheme != 'ivo'):
        raise ValueError('Invalid resourceID: {}'.format(resource_id))
    return resource_id


###############################################################################
# Common command line options and customized format
###############################################################################

class SingleMetavarHelpFormatter(RawDescriptionHelpFormatter):
    """
    Class the customizes the argparse help formatter. It does 2 things:
        - the display of an option with short and long format is shorter
            e.g '-o, --output OUTPUT' instead of the default '-o OUTPUT,
            --output OUTPUT'
        - options are sorted in alphabetical order in command line usage
    """

    def _format_action_invocation(self, action):
        """
        Customized version of the function in HelpFormatter to shorten the
        display of option with long and short options
        :param action:
        :return:
        """
        if not action.option_strings:
            metavar, = self._metavar_formatter(action, action.dest)(1)
            return metavar

        else:
            parts = []

            # if the Optional doesn't take a value, format is:
            #    -s, --long
            if action.nargs == 0:
                parts.extend(action.option_strings)

            # if the Optional takes a value, format is:
            #    -s ARGS, --long ARGS
            else:
                default = action.dest.upper()
                args_string = self._format_args(action, default)

                parts.extend(action.option_strings)
                parts[-1] += ' %s' % args_string
            return ', '.join(parts)

    def add_arguments(self, actions):
        """
        Customized version to sort options in alphabetical order before
        displaying them
        :param actions:
        :return:
        """
        if (len(actions) > 0) and \
           (actions[0].container.title in ['optional arguments', 'options']):
            actions[0].container.title = 'optional arguments'
            actions = sorted(actions, key=attrgetter('dest'))
        super(SingleMetavarHelpFormatter, self).add_arguments(actions)


class _AugmentAction(object):
    """
    This automatically adds parents and the formatter class when a new
    subparser is created in the client code
    """

    def __init__(self, parser, action):
        self.action = action
        self.parser = parser

    def add_parser(self, name, **kwargs):
        kwargs['parents'] = [self.parser]
        kwargs['formatter_class'] = SingleMetavarHelpFormatter
        return self.action.add_parser(name, **kwargs)


class _CustomArgParser(ArgumentParser):
    """
    Custom arg parses to sort options in alphabetical order before displaying
    them
    """

    def __init__(self, subparsers=True, common_parser=None, version=None,
                 **kwargs):
        self.common_parser = common_parser
        self.subparsers = subparsers
        self._subparsers_added = False
        self.kwargs = kwargs
        self._version = version
        kwargs['formatter_class'] = SingleMetavarHelpFormatter
        if not self.subparsers:
            super(_CustomArgParser, self).__init__(parents=[common_parser],
                                                   **kwargs)
        else:
            super(_CustomArgParser, self).__init__(**kwargs)
        if version is not None:
            self.add_argument('-V', '--version', action='version',
                              version=version)

    def add_subparsers(self, **kwargs):
        if not self.subparsers:
            raise RuntimeError('Parser created to run without subparsers')
        self._subparsers_added = True
        return _AugmentAction(self.common_parser,
                              super(_CustomArgParser, self).add_subparsers(
                                  **kwargs))

    def parse_args(self, args=None, namespace=None):
        if self.subparsers and not self._subparsers_added:
            raise RuntimeError('No subparsers added. Change the parsers flag?')
        result = super(_CustomArgParser, self).parse_args(args=args,
                                                          namespace=namespace)
        if hasattr(result, 'verbose') and result.verbose:
            logging.basicConfig(level=logging.INFO, stream=sys.stdout)
            logger.setLevel(logging.INFO)
        elif hasattr(result, 'debug') and result.debug:
            logging.basicConfig(level=logging.DEBUG, stream=sys.stdout)
            logger.setLevel(logging.DEBUG)
        elif hasattr(result, 'quiet') and result.quiet:
            logging.basicConfig(level=logging.FATAL, stream=sys.stdout)
            logger.setLevel(logging.FATAL)
        else:
            logging.basicConfig(level=logging.WARNING, stream=sys.stdout)

        # print package version info
        with warnings.catch_warnings():
            warnings.simplefilter("error", category=VersionWarning)
            try:
                check_version(self._version)
            except VersionWarning as e:
                logger.warning('{}'.format(str(e)))
            except Exception as e:
                logger.debug('Unexpected exception in check_version: {}'.format(str(e)))
                pass
        return result


def get_base_parser(subparsers=True, version=None, usecert=True,
                    default_resource_id=None, auth_required=False,
                    service=None):
    """
    An ArgumentParser with some common things most CADC clients will want.
    There are two modes to use this parser: with or without subparsers.
    With supbarsers (subparsers=True), separate subparsers are created for
    each subcommand and the common CADC options show up in each subcommand
    (and not on the parent parser). Without subparsers (subparsers=False)
    the common options are automatically added to the base parser.

    :param subparsers: True if the parser will use subparsers (subcommands)
    otherwise False
    :param version: A version number if desired.
    :param usecert: If True add '--cert' argument.
    :param default_resource_id: default resource identifier to use. (deprecated
    in favour of service argument)
    :param auth_required: At least one of the authentication options is
    required
    :param service: Alias of resource_id
    :return: An ArgumentParser instance.
    """
    cparser = ArgumentParser(add_help=False,
                             formatter_class=SingleMetavarHelpFormatter)

    auth_group = cparser.add_mutually_exclusive_group(required=auth_required)
    if usecert:
        auth_group.add_argument(
            '--cert', type=str,
            help='location of your X509 certificate to use for ' +
                 'authentication (unencrypted, in PEM format)')
    auth_group.add_argument('-n', action='store_true',
                            help='use .netrc in $HOME for authentication')
    auth_group.add_argument('--netrc-file',
                            help='netrc file to use for authentication')
    auth_group.add_argument('-u', '--user',
                            help='name of user to authenticate. ' +
                                 'Note: application prompts for the '
                                 'corresponding password!')
    cparser.add_argument('--host', help=SUPPRESS)
    cparser.add_argument('-k', '--insecure', action='store_true',
                         help=SUPPRESS)
    if service is None:
        if default_resource_id is None:
            cparser.add_argument('--resource-id',
                                 type=urlparse, required=True,
                                 help='resource identifier '
                                      '(e.g. ivo://cadc.nrc.ca/service)')
        else:
            cparser.add_argument('--resource-id', type=parse_resource_id,
                                 default=default_resource_id,
                                 help='resource identifier (default {})'.
                                 format(default_resource_id))
    else:
        cparser.add_argument(
            '-s', '--service', action=_ServiceAction,
            default=service,
            help='service this command accesses. Both IDs in short '
                 'form (<service>) or the complete one '
                 '(ivo://cadc.nrc.ca/<service>) as well as actual URLs to the '
                 'root of the service (https://someurl/service) are accepted.'
                 ' Default is: {}'.format(service))
    log_group = cparser.add_mutually_exclusive_group()
    log_group.add_argument('-d', '--debug', action='store_true',
                           help=SUPPRESS)
    log_group.add_argument('-q', '--quiet', action='store_true',
                           help='run quietly')
    log_group.add_argument('-v', '--verbose', action='store_true',
                           help='verbose messages')

    argparser = _CustomArgParser(subparsers=subparsers, common_parser=cparser,
                                 version=version)
    return argparser


class _ServiceAction(Action):
    def __init__(self, option_strings, dest, nargs=None, **kwargs):
        if nargs is not None:
            raise ValueError("nargs not allowed")
        super(_ServiceAction, self).__init__(option_strings, dest, **kwargs)

    def __call__(self, parser, namespace, values, option_string=None):
        if not (values.startswith('ivo://') or (values.startswith('http'))):
            values = 'ivo://cadc.nrc.ca/{}'.format(values)
        setattr(namespace, self.dest, values)


class Md5File(object):
    """
    A wrapper to a file object that calculates the MD5 sum of the bytes
    that are being read or written. It allows allows seeking to a particular
    position in the file and it limits the legth of the data to be read,
    essentially enabling chunking/segmentation of large files.
    """

    def __init__(self, f, mode, offset=0, length=None):
        """

        :param f: location of the file
        :param mode: open mode (must be binary)
        :param offset: offset to start from
        :param length: data length to read from
        """
        self.file = open(f, mode)
        self._offset = offset
        self._length = length
        self._end_offset = os.stat(f).st_size
        if offset:
            if offset > self._end_offset:
                raise AttributeError(
                    '{} offset greater that file size: {} vs {}'.format(
                        f, offset, self._end_offset))
            self.file.seek(offset)
        if length:
            self._end_offset = min(offset + length, self._end_offset)
        self._md5_checksum = hashlib.md5()
        self._total_read_length = 0

    def __enter__(self):
        return self

    def read(self, size):
        batch_read_size = \
            min(size, self._end_offset-self._total_read_length-self._offset)
        buffer = self.file.read(batch_read_size)
        self._total_read_length += len(buffer)
        self._md5_checksum.update(buffer)
        return buffer

    def write(self, buffer):
        self._md5_checksum.update(buffer)
        self.file.write(buffer)
        self.file.flush()

    def __exit__(self, *args, **kwargs):
        if not self.file.closed:
            self.file.close()
        # clean up
        exit = getattr(self.file, '__exit__', None)
        if exit is not None:
            return exit(*args, **kwargs)
        else:
            exit = getattr(self.file, 'close',
                           None)
            if exit is not None:
                exit()

    def __getattr__(self, attr):
        if attr not in ['__len__', 'len']:
            # requests or other libraries might want to use `tell` and `seek`
            # to retry etc. Do not allow it as it might interfere with the
            # md5 checksum calculations
            raise AttributeError('Unsupported attribute: {}'.format(attr))

    def __len__(self):
        """
        :return: size meant to be seen by the clients (entire file or just
        a segment/chunk)
        """
        sz = self._end_offset - self._offset
        return sz

    def len(self):
        return self.__len__()

    def __iter__(self):
        return iter(self.file)

    @property
    def md5_checksum(self):
        return self._md5_checksum.hexdigest()


def get_url_content(url, cache_file, refresh_interval, verify=True):
    """
     Return content of a url from a cache file or directly from source if the
     cache is stale (file is older than refresh_interval). For now, the access
     to the url is anonymous.
     :param url: URL of the source
     :param cache_file: cache file location
     :param refresh_interval: how long (in sec) to consider the cache stale
     :param verify: verify the HTTPS server certificate. This should be
     set to False only for testing purposes

     :returns content of the url from source or cache
    """
    content = None
    if os.path.exists(cache_file):
        last_accessed = os.path.getmtime(cache_file)
    else:
        last_accessed = None
    if (last_accessed and (time.time() - last_accessed) < refresh_interval):
        # get content from the cached file
        logger.debug(
            'Read cached content of {}'.format(cache_file))
        try:
            with open(cache_file, 'r') as f:
                content = f.read()
        except Exception:
            # will download it
            pass
    # config dirs if they don't exist yet
    if not os.path.exists(os.path.dirname(cache_file)):
        os.makedirs(os.path.dirname(cache_file))

    if not content:
        # get content from source URL
        try:
            session = requests.Session()
            # do not allow requests to use .netrc file
            session.trust_env = False
            rsp = session.get(url, verify=verify)
            rsp.raise_for_status()
            content = rsp.text
            if content is None or len(content.strip(' ')) == 0:
                # workaround for a problem with CADC servers
                raise exceptions.HttpException('Received empty content')
            with open(cache_file, 'w') as f:
                f.write(content)
        except Exception as e:
            # problems with the source. Try to use the old
            # local one regardless of how old it is
            logger.debug("Cannot read content from {}: {}".format(url, str(e)))
            if os.path.exists(cache_file):
                with open(cache_file, 'r') as f:
                    content = f.read()
                # touch the cache file to reduce the remote calls to the frequency dictated
                # by the refresh_interval even if errors are later encountered.
                Path(cache_file).touch()
    if content is None:
        raise RuntimeError('Cannot get content from {}'.format(url))
    return content


class VersionWarning(Warning):
    # category of warnings when current version falls behind the released version
    pass


def check_version(version):
    """
    Function to test the version of a package against PyPI and issue
    a warning when a newer version exists on PyPI. It will check only against
    versions of format "major.minor.micro" and ignore the others including pre-releases
    or dev releases.

    Note: Only the first check of a package is performed. Subsequent calls
    for the same package are ignored.

    :param version: package version of the form "<package_name> <version>
    e.g. "cadcutils 1.2.3"

    :return: raises a VersionWarning when PyPI version is ahead.
    """
    strict_versions = None
    current_version = None
    try:
        package, pkg_version = version.split(' ')
        if package in check_version.checked:
            return
        check_version.checked.append(package)
        cache_file = os.path.join(CADC_CACHE_DIR, package, 'caches/.pypi_versions.json')
        content = get_url_content(
            url='https://pypi.org/pypi/{}/json'.format(package),
            cache_file=cache_file,
            refresh_interval=VERSION_REFRESH_INTERVAL)
        data = json.loads(content)
        versions = list(data['releases'].keys())
        current_version = Version(pkg_version)
        strict_versions = []
        for v in versions:
            try:
                tempv = Version(v)  # version
                if not tempv.is_devrelease and not tempv.is_prerelease:
                    strict_versions.append(
                        Version('{}.{}.{}'.format(tempv.major, tempv.minor, tempv.micro)))
            except ValueError:
                continue
        strict_versions.sort()
    except Exception as e:
        logger.debug(
            'Unexpected exception in PyPI version checking: {}'.format(str(e)))
        pass
    if strict_versions and current_version < strict_versions[-1]:
        warnings.warn('Current version {}. A newer version, {}, '
                      'is available on PyPI'.format(version, strict_versions[-1]),
                      category=VersionWarning)


check_version.checked = []  # packages already checked



# -*- coding: utf-8 -*-
# ***********************************************************************
# ******************  CANADIAN ASTRONOMY DATA CENTRE  *******************
# *************  CENTRE CANADIEN DE DONNÉES ASTRONOMIQUES  **************
#
#  (c) 2023.                            (c) 2023.
#  Government of Canada                 Gouvernement du Canada
#  National Research Council            Conseil national de recherches
#  Ottawa, Canada, K1A 0R6              Ottawa, Canada, K1A 0R6
#  All rights reserved                  Tous droits réservés
#
#  NRC disclaims any warranties,        Le CNRC dénie toute garantie
#  expressed, implied, or               énoncée, implicite ou légale,
#  statutory, of any kind with          de quelque nature que ce
#  respect to the software,             soit, concernant le logiciel,
#  including without limitation         y compris sans restriction
#  any warranty of merchantability      toute garantie de valeur
#  or fitness for a particular          marchande ou de pertinence
#  purpose. NRC shall not be            pour un usage particulier.
#  liable in any event for any          Le CNRC ne pourra en aucun cas
#  damages, whether direct or           être tenu responsable de tout
#  indirect, special or general,        dommage, direct ou indirect,
#  consequential or incidental,         particulier ou général,
#  arising from the use of the          accessoire ou fortuit, résultant
#  software.  Neither the name          de l'utilisation du logiciel. Ni
#  of the National Research             le nom du Conseil National de
#  Council of Canada nor the            Recherches du Canada ni les noms
#  names of its contributors may        de ses  participants ne peuvent
#  be used to endorse or promote        être utilisés pour approuver ou
#  products derived from this           promouvoir les produits dérivés
#  software without specific prior      de ce logiciel sans autorisation
#  written permission.                  préalable et particulière
#                                       par écrit.
#
#  This file is part of the             Ce fichier fait partie du projet
#  OpenCADC project.                    OpenCADC.
#
#  OpenCADC is free software:           OpenCADC est un logiciel libre ;
#  you can redistribute it and/or       vous pouvez le redistribuer ou le
#  modify it under the terms of         modifier suivant les termes de
#  the GNU Affero General Public        la “GNU Affero General Public
#  License as published by the          License” telle que publiée
#  Free Software Foundation,            par la Free Software Foundation
#  either version 3 of the              : soit la version 3 de cette
#  License, or (at your option)         licence, soit (à votre gré)
#  any later version.                   toute version ultérieure.
#
#  OpenCADC is distributed in the       OpenCADC est distribué
#  hope that it will be useful,         dans l’espoir qu’il vous
#  but WITHOUT ANY WARRANTY;            sera utile, mais SANS AUCUNE
#  without even the implied             GARANTIE : sans même la garantie
#  warranty of MERCHANTABILITY          implicite de COMMERCIALISABILITÉ
#  or FITNESS FOR A PARTICULAR          ni d’ADÉQUATION À UN OBJECTIF
#  PURPOSE.  See the GNU Affero         PARTICULIER. Consultez la Licence
#  General Public License for           Générale Publique GNU Affero
#  more details.                        pour plus de détails.
#
#  You should have received             Vous devriez avoir reçu une
#  a copy of the GNU Affero             copie de la Licence Générale
#  General Public License along         Publique GNU Affero avec
#  with OpenCADC.  If not, see          OpenCADC ; si ce n’est
#  <http://www.gnu.org/licenses/>.      pas le cas, consultez :
#                                       <http://www.gnu.org/licenses/>.
#
#
# ***********************************************************************

import pytest
from unittest.mock import patch, Mock, MagicMock, call
from requests.utils import CaseInsensitiveDict
import base64

from cadcutils.net import get_header_filename, extract_md5, Transfer, \
    add_md5_header


def test_get_header_filename():
    assert 'example.txt' == get_header_filename(
        {'content-disposition': 'Attachment; filename=example.txt'})

    assert 'an example.txt' == get_header_filename(
        {'content-disposition': 'INLINE; FILENAME= "an example.txt" '})

    assert '€ rates.txt' == get_header_filename(
        {'content-disposition':
         "attachment;filename*= UTF-8''%e2%82%ac%20rates.txt"})

    assert '€ rates.txt' == get_header_filename(
        {'content-disposition': "attachment; filename=\"EURO rates.txt\";"
                                "filename*=utf-8''%e2%82%ac%20rates.txt"})


def test_md5_headers():
    md5_checksum = 'd41d8cd98f00b204e9800998ecf8427e'
    headers = CaseInsensitiveDict()
    add_md5_header(headers=headers, md5_checksum=md5_checksum)
    assert md5_checksum == extract_md5(headers=headers)
    # ensure that only the 16bytes of the md5 are sent and not the 32 of hex representation
    assert 24 == len(headers['digest'][4:])  # 16bytes->24 base64 chars

    # temporary backwards compatibility (to be removed later)
    headers['digest'] = 'md5=' + base64.b64encode(md5_checksum.encode('ascii')).decode('ascii')
    # send 32 bytes representation of the hex md5 value (44 bytes of base64)
    assert 44 == len(headers['digest'][4:])  # 32bytes -> 44 base64 chars
    assert md5_checksum == extract_md5(headers=headers)


# patch sleep to stop the test from sleeping and slowing down execution
@patch('cadcutils.net.netutils.time.sleep', MagicMock(), create=True)
def test_transfer_error():
    session = Mock()
    testservice_url = 'https://sometestservice.server/testservice'

    session.get.side_effect = [Mock(text='COMPLETED'),
                               Mock(text='COMPLETED')]
    test_target = Transfer(session=session)

    # job successfully completed
    assert not test_target.get_transfer_error(
        testservice_url + '/results/transferDetails', 'vos://testservice')
    session.get.assert_called_with(testservice_url + '/phase',
                                   allow_redirects=True)

    # job suspended
    session.reset_mock()
    session.get.side_effect = [Mock(text='COMPLETED'),
                               Mock(text='ABORTED')]
    with pytest.raises(OSError):
        test_target.get_transfer_error(
            testservice_url + '/results/transferDetails', 'vos://testservice')
    # check arguments for session.get calls

    session.get.assert_has_calls(
        [call(testservice_url + '/phase', allow_redirects=True),
         call(testservice_url + '/phase', allow_redirects=True)])

    # job encountered an internal error
    session.reset_mock()
    session.get.side_effect = [Mock(text='COMPLETED'),
                               Mock(text='ERROR'),
                               Mock(text='InternalFault')]
    with pytest.raises(OSError):
        test_target.get_transfer_error(
            testservice_url + '/results/transferDetails', 'vos://testservice')
    session.get.assert_has_calls(
        [call(testservice_url + '/phase', allow_redirects=True),
         call(testservice_url + '/phase', allow_redirects=True),
         call(testservice_url + '/error')])

    # job encountered an unsupported link error
    session.reset_mock()
    link_file = 'testlink.fits'
    session.get.side_effect = [Mock(text='COMPLETED'),
                               Mock(text='ERROR'),
                               Mock(
                                   text="Unsupported link target: " +
                                        link_file)]
    assert link_file == test_target.get_transfer_error(
        testservice_url + '/results/transferDetails', 'vos://testservice')
    session.get.assert_has_calls(
        [call(testservice_url + '/phase', allow_redirects=True),
         call(testservice_url + '/phase', allow_redirects=True),
         call(testservice_url + '/error')])


def test_transfer():
    session = Mock()
    redirect_response = Mock()
    redirect_response.status_code = 303
    redirect_response.headers = \
        {'Location': 'https://transfer.host/transfer'}
    response = Mock()
    response.status_code = 200
    response.content = (
        '<?xml version="1.0" encoding="UTF-8"?>'
        '<vos:transfer xmlns:vos="http://www.ivoa.net/xml/VOSpace/v2.0" '
        'version="2.1">'
        '<vos:target>vos://some.host~vault/abc</vos:target>'
        '<vos:direction>pullFromVoSpace</vos:direction>'
        '<vos:protocol uri="ivo://ivoa.net/vospace/core#httpsget">'
        '<vos:endpoint>https://transfer.host/transfer/abc</vos:endpoint>'
        '<vos:securityMethod '
        'uri="ivo://ivoa.net/sso#tls-with-certificate" />'
        '</vos:protocol>'
        '<vos:keepBytes>true</vos:keepBytes>'
        '</vos:transfer>')
    session.post.return_value = redirect_response
    session.get.return_value = response
    test_transfer = Transfer(session=session)
    test_transfer.get_transfer_error = Mock()  # not transfer error
    protocols = test_transfer.transfer(
        'https://some.host/service', 'vos://abc', 'pullFromVoSpace')
    assert protocols == ['https://transfer.host/transfer/abc']

    session.reset_mock()
    session.post.return_value = Mock(status_code=404)
    with pytest.raises(OSError) as e:
        test_transfer.transfer(
            'https://some.host/service', 'vos://abc',
            'pullFromVoSpace')
        assert 'File not found: vos://abc' == str(e)

    session.reset_mock()
    session.post.return_value = Mock(status_code=500)
    with pytest.raises(OSError) as e:
        test_transfer.transfer(
            'https://some.host/service', 'vos://abc',
            'pullFromVoSpace')
        assert 'Failed to get transfer service response.' == str(e)
